#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Server module.

# File: mod_srv.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: Service class
"""

from kb4it.core.log import get_logger
from kb4it.core.util import get_traceback


class Service:
    """
    Service class.

    It is the base class for those modules acting as services.
    Different modules (GUI, Database, Ask, etc...) share same methods
    which is useful to start/stop them, simplify logging and, comunicate
    each other easily.
    """

    log = None
    logname = None
    section = None
    section_name = None

    def __init__(self, app=None):
        """Initialize Service instance."""
        if app is not None:
            self.app = app

        self.started = False

    def is_started(self):
        """
        Check if service is started.

        Return True or False if service is running / not running
        """
        return self.started

    def print_traceback(self):
        """Print traceback."""
        self.log.debug(get_traceback())

    def start(self, app, logname, section_name):
        """Start service."""
        self.started = True
        self.app = app
        self.logname = logname
        self.section_name = section_name
        params = self.app.get_params()
        severity = params.LOGLEVEL
        self.log = get_logger(logname, severity)
        self.initialize()
        # ~ self.log.debug("Module %s started" , logname)

    def end(self):
        """End service.

        Use finalize for writting a custom end method
        """
        self.started = False
        self.finalize()

    def initialize(self):
        """Initialize service.

        All clases derived from Service class must implement this method
        """
        self.log.debug("Service %s started", self.logname)

    def finalize(self):
        """Finalize service.

        All clases derived from Service class must implement this method
        """
        self.log.debug("Service %s finished", self.logname)

    def get_service(self, name):
        """Get service name."""
        return self.app.get_service(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
RDF Graph In Memory database module.

# Author: Tomás Vírseda <tomasvirseda@gmail.com>
# License: GPLv3
# Description: In-memory database module
"""


from kb4it.core.service import Service
from kb4it.core.util import guess_datetime, sort_dictionary


class KB4ITDB(Service):
    """KB4IT database class."""

    db = {}
    sorted_docs = []
    blocked_keys = []
    ignored_keys = []

    def initialize(self):
        """Initialize database module."""
        params = self.app.get_params()
        self.sort_attribute = params.SORT_ATTRIBUTE
        self.db = {}
        self.sorted_docs = []
        self.ignored_keys = self.blocked_keys = ['Title', 'Timestamp']

    def del_document(self, doc):
        """Delete a document node from database."""
        adoc = "%s.adoc" % doc
        try:
            del(self.db[adoc])
            self.log.debug("DOC[%s] deleted from database", doc)
            self.sort_database()
        except KeyError:
            self.log.debug("DOC[%s] not found in database", doc)

    def add_document(self, doc):
        """Add a new document node to the database."""
        self.db[doc] = {}
        self.log.debug("* DOC[%s] added to database", doc)

    def add_document_key(self, doc, key, value):
        """Add a new key/value node for a given document."""
        try:
            alist = self.db[doc][key]
            alist.append(value)
            self.db[doc][key] = alist
        except KeyError:
            self.db[doc][key] = [value]

        self.log.debug("* DOC[%s] KEY[%s] VALUE[%s] added", doc, key, value)

    def get_blocked_keys(self):
        """Return blocked keys"""
        return self.blocked_keys

    def get_ignored_keys(self):
        """Return ignored keys"""
        return self.ignored_keys

    def ignore_key(self, key):
        """Add given key to ignored keys list"""
        self.ignored_keys.append(key)

    def sort_database(self):
        """
        Build a list of documents sorted by the given date attribute
        in descending order
        """
        self.sorted_docs = self.sort_by_date(list(self.db.keys()))

    def sort_by_date(self, doclist):
        """Build a list of documents sorted by timestamp desc."""
        sorted_docs = []
        adict = {}
        for doc in doclist:
            sdate = self.get_doc_timestamp(doc)
            ts = guess_datetime(sdate)
            if ts is not None:
                adict[doc] = ts
            else:
                self.log.error("Doc '%s' doesn't have a valid timestamp?", doc)
        alist = sort_dictionary(adict)
        for doc, timestamp in alist:
            sorted_docs.append(doc)
        return sorted_docs

    def get_documents(self):
        """
        Return a list of docs sorted by date (timestamp or sort
        attribute.
        """
        return self.sorted_docs

    def get_doc_timestamp(self, doc):
        """Get timestamp for a given document."""
        try:
            timestamp = self.db[doc][self.sort_attribute][0]
        except:
            timestamp = self.db[doc]['Timestamp'][0]
        return timestamp

    def get_doc_properties(self, doc):
        """Return a dictionary with the properties of a given doc"""
        return self.db[doc]

    def get_values(self, doc, key):
        """Return a list of values given a document and a key."""
        try:
            return self.db[doc][key]
        except KeyError:
            return ['']

    def get_all_values_for_key(self, key):
        """
        Return a list of all values for a given key sorted
        alphabetically.
        """
        values = []
        for doc in self.db:
            try:
                values.extend(self.db[doc][key])
            except KeyError:
                pass
        values = list(set(values))
        values.sort(key=lambda y: y.lower())
        return values

    def get_custom_keys(self, doc):
        """Return a list of custom keys sorted alphabetically."""
        custom_keys = []
        keys = self.get_doc_keys(doc)
        for key in keys:
            if key not in self.ignored_keys:
                custom_keys.append(key)
        custom_keys.sort(key=lambda y: y.lower())
        return custom_keys

    def get_all_keys(self):
        """Return all keys in the database sorted alphabetically."""
        blocked_keys = self.get_blocked_keys()
        keys = set()
        database = self.get_documents()
        for doc in database:
            for key in self.get_doc_keys(doc):
                if key not in blocked_keys:
                    keys.add(key)
        keys = list(keys)
        keys.sort(key=lambda y: y.lower())
        return keys

    def get_docs_by_key_value(self, key, value):
        """Return a list documents for a given key/value sorted by date"""
        docs = []
        for doc in self.db:
            try:
                if value in self.db[doc][key]:
                    docs.append(doc)
            except KeyError:
                pass
        return self.sort_by_date(docs)

    def get_doc_keys(self, doc):
        """Return a list of keys for a given doc sorted alphabetically."""
        keys = []
        for key in self.db[doc]:
            keys.append(key)
        keys.sort(key=lambda y: y.lower())
        return keys



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Server module.

# File: theme.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: techdoc theme scripts
"""

from kb4it.services.builder import KB4ITBuilder

class Theme(KB4ITBuilder):
    def hello(self):
        self.log.debug("This is the theme techdoc")

    def generate_sources(self):
        pass

    def build(self):
        # Default pages
        self.create_page_properties()
        self.create_page_stats()
        self.create_page_index_all()
        self.create_page_index()
        self.create_page_about_app()
        self.create_page_about_theme()
        self.create_page_about_kb4it()
        self.create_page_help()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Custom theme module.

# File: theme.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: newspaper theme scripts
"""

import os
import sys
import glob
import uuid
from urllib.parse import urlparse
from datetime import datetime, timedelta
from concurrent.futures import ThreadPoolExecutor as Executor

import nltk
from nltk.tokenize import sent_tokenize, word_tokenize
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
from dateutil.parser import parse
import feedparser

from kb4it.services.builder import KB4ITBuilder
from kb4it.core.util import valid_filename, guess_datetime
from kb4it.core.util import sort_dictionary

IGNORE = set()
for token in tuple(nltk.corpus.stopwords.words('french')):
    IGNORE.add(token.lower())

for token in tuple(nltk.corpus.stopwords.words('english')):
    IGNORE.add(token.lower())

for token in tuple(nltk.corpus.stopwords.words('spanish')):
    IGNORE.add(token.lower())

tokens = list(IGNORE)
tokens.sort()


def get_tags(text):
    # word stemming and polishing
    mylist = set()
    ps = PorterStemmer()
    words = word_tokenize(text)
    for word in words:
        stem_word = ps.stem(word)
        if stem_word not in words:
            stem_word = words[0]
        mylist.add(stem_word)

    # FIlter tags according basic rules
    stags = set()
    for w in mylist:
        if w.lower() not in IGNORE:
            if w.startswith('LG_'):
                stags.add(w)
            else:
                if not w.isdigit():
                    if w.isalnum() and len(w) > 3:
                        stags.add(w.lower())
    ltags = list(stags)
    ltags.sort()
    return ltags

class Theme(KB4ITBuilder):
    feeds = {}

    def clean_source_dir(self):
        source_path = self.srvapp.get_source_path()
        adocs = glob.glob(os.path.join(source_path, "*.adoc"))
        for adoc in adocs:
            # ~ self.log.debug("Deleting %s", adoc)
            os.unlink(adoc)

    def generate_sources(self):
        self.log.debug("\t\tClean source directory from previous executions")
        self.clean_source_dir()

        self.log.info("\t\tCreating sources from rss")
        source_path = self.srvapp.get_source_path()
        rss_file = os.path.join(source_path, 'rss.txt')
        if os.path.exists(rss_file):
            sources = open(rss_file, 'r').read().splitlines()
            self.process_feeds(sources)
            self.validate_rss_feeds()
            self.create_newspapers_page()
        else:
            self.log.error ("RSS file path doesn't exist. Exit.")
            exit(-1)

    def build(self):
        # Default pages
        self.create_all_keys_page()
        self.create_properties_page()
        self.create_stats_page()
        self.create_index_all()
        self.create_index_page()
        self.create_about_page()
        self.create_help_page()

    def get_doc_card_np_entry_short(self, rss, eid):
        try:
            CARD_NP_ENTRY = self.template('CARD_DOC_NEWSPAPER_ENTRY_SHORT')
            # ~ thumbnail = self.feeds[rss]['entries'][eid]['thumbnail']
            title = self.feeds[rss]['entries'][eid]['title']
            summary = self.feeds[rss]['entries'][eid]['summary']
            updated = self.feeds[rss]['entries'][eid]['updated']
            link = self.feeds[rss]['entries'][eid]['link']
            thumbnail = self.get_favicon_url(link)
            tags = self.feeds[rss]['entries'][eid]['tags']
            card = CARD_NP_ENTRY % (thumbnail, link, title, updated)
            # ~ self.log.debug(card)
            return card
        except:
            # ~ self.log.error("RSS[%s] - EID[%s]", rss, eid)
            # ~ self.log.error(self.feeds[rss]['entries'])
            return ''

    def get_news(self, dtu, dtl):
        NEWS = ''
        d = {}
        for rss in self.feeds:
            for eid in self.feeds[rss]['entries']:
                updated = self.feeds[rss]['entries'][eid]['updated']
                dt = guess_datetime(updated)
                if dt is not None:
                    if dt > dtl and dt < dtu:
                        key = "%s#-#%s" % (rss, eid)
                        d[key] = dt

        alist = sort_dictionary(d)
        for key, updated in alist:
            sep = key.find('#-#')
            rss = key[:sep]
            eid = key[sep+3:]
            NEWS += self.get_doc_card_np_entry_short(rss, eid)

        if len(NEWS) == 0:
            NEWS = """<div class="uk-text-lead">No news available</div>"""

        return NEWS

    def create_index_page(self):
        PAGE_INDEX = self.template('PAGE_INDEX')

        ## TODAY ###
        adate = datetime.now()
        upper = "%4d-%02d-%02d 23:59:59" % (adate.year, adate.month, adate.day)
        lower = "%4d-%02d-%02d 00:00:00" % (adate.year, adate.month, adate.day)
        dtu = datetime.strptime(upper, "%Y-%m-%d %H:%M:%S")
        dtl = datetime.strptime(lower, "%Y-%m-%d %H:%M:%S")

        NEWS_TODAY = self.get_news(dtu, dtl)
        PAGE_TODAY = self.template('PAGE_NEWS')
        self.distribute('today', PAGE_TODAY % ('Today', NEWS_TODAY))

        ## YESTERDAY ##
        adate = datetime.today() - timedelta(days=1)
        upper = "%4d-%02d-%02d 23:59:59" % (adate.year, adate.month, adate.day)
        lower = "%4d-%02d-%02d 00:00:00" % (adate.year, adate.month, adate.day)
        dtu = datetime.strptime(upper, "%Y-%m-%d %H:%M:%S")
        dtl = datetime.strptime(lower, "%Y-%m-%d %H:%M:%S")

        NEWS_YESTERDAY = self.get_news(dtu, dtl)
        PAGE_YESTERDAY = self.template('PAGE_NEWS')
        self.distribute('yesterday', PAGE_YESTERDAY % ('Yesterday', NEWS_YESTERDAY))

        ### THIS WEEK ###
        now = datetime.now()
        adate = datetime.today() - timedelta(days=datetime.today().isoweekday() % 7)
        upper = "%4d-%02d-%02d 23:59:59" % (now.year, now.month, now.day)
        lower = "%4d-%02d-%02d 00:00:00" % (adate.year, adate.month, adate.day)
        dtu = datetime.strptime(upper, "%Y-%m-%d %H:%M:%S")
        dtl = datetime.strptime(lower, "%Y-%m-%d %H:%M:%S")

        NEWS_WEEK = self.get_news(dtu, dtl)
        PAGE_WEEK = self.template('PAGE_NEWS')
        self.distribute('week', PAGE_WEEK % ('This week', NEWS_WEEK))

        ### THIS MONTH ###
        adate = datetime.now()
        upper = "%4d-%02d-%02d 23:59:59" % (now.year, now.month, now.day)
        lower = "%4d-%02d-01 00:00:00" % (adate.year, adate.month)
        dtu = datetime.strptime(upper, "%Y-%m-%d %H:%M:%S")
        dtl = datetime.strptime(lower, "%Y-%m-%d %H:%M:%S")
        NEWS_MONTH = self.get_news(dtu, dtl)
        PAGE_MONTH = self.template('PAGE_NEWS')
        self.distribute('month', PAGE_MONTH % ('This month', NEWS_MONTH))

        ### PAST MONTH ###
        today = datetime.today()
        first = today.replace(day=1)
        pastmonth = first - timedelta(days=1)
        firstpast = pastmonth.replace(day=1)
        upper = "%4d-%02d-%02d 23:59:59" % (pastmonth.year, pastmonth.month, pastmonth.day)
        lower = "%4d-%02d-%02d 00:00:00" % (firstpast.year, firstpast.month, firstpast.day)
        dtu = datetime.strptime(upper, "%Y-%m-%d %H:%M:%S")
        dtl = datetime.strptime(lower, "%Y-%m-%d %H:%M:%S")
        NEWS_PAST = self.get_news(dtu, dtl)
        PAGE_PAST = self.template('PAGE_NEWS')
        self.distribute('past', PAGE_PAST % ('Past month', NEWS_PAST))

        # ~ self.distribute_to_source('index', PAGE_INDEX % (NEWS_TODAY, NEWS_YESTERDAY, NEWS_WEEK))
        self.distribute('index', PAGE_INDEX % (NEWS_TODAY, NEWS_YESTERDAY, NEWS_WEEK, NEWS_MONTH, NEWS_PAST))

    def get_favicon_url(self, link):
        o = urlparse(link)
        rss_scheme = o.scheme
        rss_host = o.netloc
        favicon = "https://www.google.com/s2/favicons?domain=%s" % o.netloc
        # ~ favicon = "%s://%s/favicon.ico" % (rss_scheme, rss_host)
        return favicon

    def get_homepage(self, link):
        o = urlparse(link)
        rss_scheme = o.scheme
        rss_host = o.netloc
        homepage = "%s://%s" % (rss_scheme, rss_host)
        return homepage

    def get_website(self, link):
        o = urlparse(link)
        return o.netloc

    def create_newspaper_page(self, rss):
        PAGE_NP = self.template('PAGE_NEWSPAPER')
        PAGE_NP_ENTRY = self.template('PAGE_NEWSPAPER_ENTRY')
        CARD_NP_ENTRY = self.template('CARD_DOC_NEWSPAPER_ENTRY')
        rss_title = self.feeds[rss]['header']['title']
        rss_summary = self.feeds[rss]['header']['summary']
        rss_link = self.feeds[rss]['header']['link']
        # ~ rss_updated = self.feeds[rss]['header']['updated']

        rss_properties = ''
        # ~ rss_properties += ':Updated: %s\n' % rss_updated
        rss_favicon = self.get_favicon_url(rss_link)

        NEWS = ''
        for eid in self.feeds[rss]['entries']:
            thumbnail = self.feeds[rss]['entries'][eid]['thumbnail']
            title = self.feeds[rss]['entries'][eid]['title']
            summary = self.feeds[rss]['entries'][eid]['summary']
            updated = self.feeds[rss]['entries'][eid]['updated']
            link = self.feeds[rss]['entries'][eid]['link']
            tags = self.feeds[rss]['entries'][eid]['tags']
            NEWS += CARD_NP_ENTRY % (thumbnail, link, title, updated, summary)

            # RSS ENTRY PAGE
            properties = ''
            properties += ':RSSFeed: %s\n' % rss # RSS
            properties += ':Website: %s\n' % self.get_website(rss_link) # Website
            # ~ properties += ':Tag: %s\n' % tags # Tags
            properties += ':Updated: %s\n' % updated

            PAGE_NAME_ENTRY = self.get_page_id()
            PAGE_CONTENT_ENTRY = PAGE_NP_ENTRY % (title, properties, thumbnail, link, title, updated, summary)
            self.distribute_to_source(PAGE_NAME_ENTRY, PAGE_CONTENT_ENTRY)

        PAGE_NAME = self.feeds[rss]['header']['pageid']
        PAGE_CONTENT = PAGE_NP % (rss_title, rss_properties, rss_favicon, rss_link, rss_title, rss_summary, NEWS)
        self.distribute_to_source(PAGE_NAME, PAGE_CONTENT)

    def create_newspapers_page(self):
        CARD = self.template('CARD_DOC_NEWSPAPER')
        PAGE_NPS = self.template('PAGE_NEWSPAPERS')
        CARDS = ''
        for rss in self.feeds:
            rss_is_valid = self.feeds[rss]['header']['valid']
            if rss_is_valid:
                # Create newspaper page
                self.create_newspaper_page(rss)
                self.log.info("\t\tRSS valid: %s" % rss)
                link = self.feeds[rss]['header']['link']
                title = self.feeds[rss]['header']['title']
                summary = self.feeds[rss]['header']['summary']
                favicon = self.get_favicon_url(link)
                homepage = self.get_homepage(link)
                npurl = "%s.html" % self.feeds[rss]['header']['pageid']
                CARDS += CARD % (favicon, npurl, title, summary)
            else:
                self.log.warning("RSS not valid: %s" % rss)

        # Create page for all newspapers
        PAGE_CONTENT = PAGE_NPS % CARDS
        PAGE_NAME = 'newspapers'
        self.distribute(PAGE_NAME, PAGE_CONTENT)

    def validate_field(self, rss, field):
        try:
            value = self.feeds[rss]['header'][field]
            found = True
        except:
            found = False

        # ~ self.log.debug("RSS[%s][%s] = %s", rss, field, found)
        return found

    def validate_rss_feeds(self):
        core_fields = ['title', 'link', 'updated']
        for rss in self.feeds:
            has_title = self.validate_field(rss, 'title')
            has_summary = self.validate_field(rss, 'summary')
            has_link = self.validate_field(rss, 'link')
            # ~ has_update = self.validate_field(rss, 'updated')
            if has_title and has_summary and has_link:
                self.feeds[rss]['header']['valid'] = True
            else:
                self.feeds[rss]['header']['valid'] = False

    def create_newspaper(self):
        pass
        # ~ self.log.info("\t\t<-- Creating newspaper")
        # ~ source_path = self.srvapp.get_source_path()
        # ~ rss_file = os.path.join(source_path, 'rss.txt')
        # ~ if os.path.exists(rss_file):
            # ~ sources = open(rss_file, 'r').read().splitlines()
            # ~ self.process_feeds(sources)
            # ~ self.validate_feeds()
            # ~ self.create_newspapers_page()
        # ~ else:
            # ~ self.log.error ("RSS file path doesn't exist. Exit.")
            # ~ exit(-1)
        # ~ self.log.info("\t\tNewspaper created -->")


    def get_rss_data(self, rss):
        try:
            f = feedparser.parse(rss)
            return rss, f['feed'], f['entries']
        except:
            self.log.error("Error while getting data from: %s" % rss)
            return None

    def get_chanel_info(self, channel):
        header = {}
        for field in channel:
            # Only strings
            # ~ if isinstance(channel[field], str):
                # ~ header[field] = channel[field]
            # All fields:
            header[field] = channel[field]
        return header

    def get_field_value(self, entry, field):
        try:
            return entry[field]
        except:
            return ''

    def get_page_id(self):
        return str(uuid.uuid1())
        # ~ o = urlparse(link)
        # ~ rss_host = o.netloc
        # ~ rss_path = o.path
        # ~ return "%s_%s" % (valid_filename(rss_host.replace('.', '_')), valid_filename(rss_path))

    def process_feeds(self, sources):
        self.log.info("\t\tProcessing feeds")
        fieldsets = []
        with Executor(max_workers=5) as exe:
            jobs = []
            for rss in sources:
                if rss.startswith('#'):
                    continue
                job = exe.submit(self.get_rss_data, rss)
                jobs.append(job)
                # ~ self.log.debug("\t\tQuerying feed: %s" % rss)

            for job in jobs:
                rss, channel, entries = job.result()
                # ~ self.log.debug ("\t\tReceiving data from: %s" % channel['title'])

                # Process feed
                self.feeds[rss] = {}
                fieldset = set()

                ## Process feed header
                self.feeds[rss]['header']={}
                for field in channel:
                    fieldset.add(field)
                    if isinstance(channel[field], str):
                        self.feeds[rss]['header'][field] = channel[field]
                    elif isinstance(channel[field], feedparser.FeedParserDict):
                        pass

                ### Adjust summary field
                try:
                    summary = self.feeds[rss]['header']['subtitle']
                    self.feeds[rss]['header']['summary'] = summary
                    del(self.feeds[rss]['header']['subtitle'])
                except Exception as error:
                    self.log.error(error)
                    pass #sys.exit()

                ### Generate page id
                try:
                    pageid = self.get_page_id()#self.get_page_id(self.feeds[rss]['header']['link'])
                    self.feeds[rss]['header']['pageid'] = pageid
                except Exception as error:
                    self.log.error(error)
                    sys.exit()

                # ~ self.log.debug(self.feeds[rss]['header'])
                fieldsets.append(fieldset)

                ## Process feed entries
                self.feeds[rss]['entries']={}
                self.log.debug("RSS[%s]: processing entries", rss)
                for entry in entries:
                    eid = self.get_page_id() #entry['id']
                    try:
                        self.feeds[rss]['entries'][eid] = {}
                        self.feeds[rss]['entries'][eid]['RSSFeed'] = rss
                        self.feeds[rss]['entries'][eid]['title'] = self.get_field_value(entry, 'title')
                        self.feeds[rss]['entries'][eid]['summary'] = self.get_field_summary(entry)
                        self.feeds[rss]['entries'][eid]['thumbnail'] = self.get_field_value(entry, 'thumbnail')
                        self.feeds[rss]['entries'][eid]['link'] = self.get_field_value(entry, 'link')
                        self.feeds[rss]['entries'][eid]['updated'] = self.get_field_updated(entry)
                        tags = get_tags(self.feeds[rss]['entries'][eid]['title'])
                        self.feeds[rss]['entries'][eid]['tags'] = ', '.join(tags)
                        self.log.debug("RSS[%s]: Entry %s added", rss, eid)
                    except Exception as error:
                        self.log.warning("Error while parsing entry for RSS: %s", rss)
                        del(self.feeds[rss]['entries'][eid])


    def get_field_summary(self, entry):
        found = False
        fields = ['description', 'subtitle', 'summary']
        for field in entry:
            # ~ self.log.debug("Checking field: %s", field)
            if field in fields:

                try:
                    summary = entry[field]
                    found = True
                    break;
                except Exception as error:
                    self.log.error(", ".join(field for field in entry))

        if found:
            return summary
        else:
            return ''

    def get_field_updated(self, entry):
        found = False
        date_fields = ['updated', 'published', 'date', 'pubDate']
        for field in entry:
            if field in date_fields:
                try:
                    updated = entry[field]
                    found = True
                    break;
                except Exception as error:
                    self.log.error(", ".join(field for field in entry))

        if found:
            dt = parse(updated)
            return dt.strftime("%Y-%m-%d %H:%M:%S")
        else:
            return None

        # ~ common_fields = set.intersection(*fieldsets)
        # ~ print ("Common fields for all RSS feeds: ")
        # ~ print(common_fields)
        # ~ {'links', 'subtitle', 'title', 'title_detail', 'subtitle_detail', 'link'}
        # ~ for rss in self.feeds:
            # ~ pageid = self.feeds[rss]['header']['pageid']
            # ~ self.distribute(pageid, str(self.feeds[rss]))
            # ~ self.log.debug("%s (%s)" % (self.feeds[rss]['header']['title'], self.feeds[rss]['header']['link']))
            # ~ for eid in self.feeds[rss]['entries']:
                # ~ title = self.feeds[rss]['entries'][eid]['title']
                # ~ published = self.feeds[rss]['entries'][eid]['published']
                # ~ link = self.feeds[rss]['entries'][eid]['link']
                # ~ self.log.debug("\t* %s (%s) via %s" % (title, published, link))
            # ~ print ("\t%s" % feeds[rss]['header']['subtitle'])
            # ~ print ("\t%s" % feeds[rss]['header']['link'])
            # ~ print ("")
        # ~ print(self.feeds)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Builder service.

# File: srv_builder.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: Create KB4IT pages
"""

import os
import sys
import math
import time
import shutil
import random
import threading
from datetime import datetime
from kb4it.core.env import APP, GPATH
from kb4it.core.service import Service
from kb4it.core.util import valid_filename, guess_datetime
from kb4it.core.util import get_human_datetime, fuzzy_date_from_timestamp
from kb4it.core.util import set_max_frequency, get_font_size
from kb4it.core.util import delete_files, extract_toc

TEMPLATES = {}

class KB4ITBuilder(Service):
    """Build HTML blocks"""

    tmpdir = None
    srvdtb = None
    srvapp = None
    temp_sources = []

    def initialize(self):
        """Initialize Builder class."""
        self.get_services()
        self.tmpdir = self.srvapp.get_temp_path()
        self.srcdir = self.srvapp.get_source_path()

    def finalize(self):
        # Delete temporary sources generated by themes
        delete_files(self.temp_sources)

    def get_services(self):
        """Get services."""
        self.srvdtb = self.get_service('DB')
        self.srvapp = self.get_service('App')


    def distribute(self, name, content):
        """
        Distribute source file to temporary directory.
        Use this method when the source asciidoctor file doesn't have to
        be analyzed.
        """
        PAGE_NAME = "%s.adoc" % name
        PAGE_PATH = os.path.join(self.tmpdir, PAGE_NAME)
        with open(PAGE_PATH, 'w') as fpag:
            fpag.write(content)
        self.log.debug("PAGE[%s] distributed to temporary path", name)

    def distribute_to_source(self, name, content):
        """
        Distribute source file to user source directory.
        Use this method when the source asciidoctor file has to
        be analyzed to extract its properties.
        File path reference will be saved and deleted at the end of the
        execution.
        """
        PAGE_NAME = "%s.adoc" % name
        PAGE_PATH = os.path.join(self.srcdir, PAGE_NAME)
        self.temp_sources.append(PAGE_PATH)
        with open(PAGE_PATH, 'w') as fpag:
            fpag.write(content)
            self.log.debug("PAGE[%s] distributed to source path", name)

    def template(self, template):
        """Return the template content from default theme or user theme"""

        properties = self.srvapp.get_runtime_properties()
        theme = properties['theme']
        current_theme = theme['id']

        # Try to get the template from cache
        try:
            return TEMPLATES[template]
        except KeyError:
            # Get template from custom theme
            # If not found, get it from default theme
            template_path = os.path.join(theme['templates'], "%s.tpl" % template)
            if os.path.exists(template_path):
                self.log.debug("THEME[%s] TPL[%s] loaded", theme['id'], template)
            else:
                current_theme = 'default'
                theme_default = os.path.join(GPATH['THEMES'], os.path.join('default', 'templates'))
                template_path = os.path.join(theme_default, "%s.tpl" % template)

                if os.path.exists(template_path):
                    self.log.debug("THEME['default'] TPL[%s] loaded", template)
                else:
                    self.log.error("TPL[%s] not found. Exit.", template)
                    exit()

            # If template found, add it to cache. Otherwise, exit.
            try:
                TEMPLATES[template] = open(template_path, 'r').read()
                return TEMPLATES[template]
            except FileNotFoundError as error:
                self.log.error(error)
                exit()

    def build_page(self, future):
        """
        Compile page from asciidoctor to HTML.
        """
        THEME_ID = self.srvapp.get_theme_property('id')
        HTML_HEADER_COMMON = self.template('HTML_HEADER_COMMON')
        HTML_HEADER_DOC = self.template('HTML_HEADER_DOC')
        HTML_HEADER_NODOC = self.template('HTML_HEADER_NODOC')
        now = datetime.now()
        timestamp = get_human_datetime(now)
        time.sleep(random.random())
        x = future.result()
        cur_thread = threading.current_thread().name
        if cur_thread != x:
            adoc, rc, j = x
            # Add header and footer to compiled doc
            htmldoc = adoc.replace('.adoc', '.html')
            if os.path.exists(htmldoc):
                adoc_title = open(adoc).readlines()[0]
                title = adoc_title[2:-1]
                htmldoctmp = "%s.tmp" % htmldoc
                shutil.move(htmldoc, htmldoctmp)
                source = open(htmldoctmp, 'r').read()
                toc = extract_toc(source)
                content = self.srvapp.apply_transformations(source)
                try:
                    if 'Metadata' in content:
                        content = highlight_metadata_section(content)
                except NameError as error:
                    # Sometimes, weird links in asciidoctor sources
                    # provoke compilation errors
                    basename = os.path.basename(adoc)
                    self.log.error("ERROR!! Please, check source document '%s'.", basename)
                    self.log.error("ERROR!! It didn't compile successfully. Usually, it is because of malformed urls.")
                finally:
                    # Some pages don't have toc section. Ignore it.
                    pass

                with open(htmldoc, 'w') as fhtm:
                    len_toc = len(toc)
                    if len_toc > 0:
                        TOC = self.template('HTML_HEADER_MENU_CONTENTS_ENABLED') % toc
                    else:
                        TOC = self.template('HTML_HEADER_MENU_CONTENTS_DISABLED')
                    docname = os.path.basename(adoc)
                    # ~ docsdir = os.path.join(self.srvapp.get_source_path(), 'sources')
                    userdoc = os.path.join(os.path.join(self.srvapp.get_source_path(), docname))
                    if os.path.exists(userdoc):
                        source_code = open(userdoc, 'r').read()
                        self.srvthm = self.get_service('Theme')
                        meta_section = self.srvthm.create_metadata_section(docname)
                        PAGE = HTML_HEADER_COMMON % (title, THEME_ID, TOC) + HTML_HEADER_DOC % (title, meta_section, docname, source_code)
                        fhtm.write(PAGE)
                    else:
                        PAGE = HTML_HEADER_COMMON % (title, THEME_ID, TOC) + HTML_HEADER_NODOC % (title)
                        fhtm.write(PAGE)
                    fhtm.write(content)
                    HTML_FOOTER = self.template('HTML_FOOTER')
                    fhtm.write(HTML_FOOTER % timestamp)
                os.remove(htmldoctmp)
                return x

    # ~ def build_pagination(self, basename, doclist, optional_title=None, custom_function='build_cardset', custom_pagination_template='PAGE_PAGINATION_HEAD', fake=False):
    def build_pagination(self, pagination):
        """
        Create a page with documents.
        If amount of documents is greater than 100, split it in several pages
        """
        PG_HEAD = self.template(pagination['template'])
        PG_CARD = self.template('CARD_PAGINATION')
        num_rel_docs = len(pagination['doclist'])
        pagelist = []
        if num_rel_docs < 100:
            total_pages = 1
            k = math.ceil(num_rel_docs/total_pages)
        else:
            k = 12
            total_pages = math.ceil(num_rel_docs/k)
            if total_pages > 10:
                total_pages = 10
                k = math.ceil(num_rel_docs/total_pages)
                row = k % 3 # Always display rows with 3 elements at list
                while row != 0:
                    k += 1
                    row = k % 3
                total_pages = math.ceil(num_rel_docs/k)
            elif total_pages == 0:
                total_pages = 1
                k = math.ceil(num_rel_docs/total_pages)

        for current_page in range(total_pages):
            PAGINATION = self.template('PAGINATION_START')
            if total_pages > 0:
                for i in range(total_pages):
                    start = k*i # lower limit
                    end = k*i + k # upper limit
                    if i == current_page:
                        if total_pages - 1 == 0:
                            PAGINATION += self.template('PAGINATION_NONE')
                        else:
                            PAGINATION += self.template('PAGINATION_PAGE_ACTIVE') % (i, start, end, num_rel_docs,i)
                        cstart = start
                        cend = end
                    else:
                        if i == 0:
                            PAGE = "%s.adoc" % pagination['basename']
                        else:
                            PAGE = "%s-%d.adoc" % (pagination['basename'], i)
                        PAGINATION += self.template('PAGINATION_PAGE_INACTIVE') % (i, start, end, num_rel_docs, PAGE.replace('adoc','html'), i)
            PAGINATION += self.template('PAGINATION_END')

            if current_page == 0:
                name = "%s" % pagination['basename']
            else:
                name = "%s-%d" % (pagination['basename'], current_page)
            ps = cstart
            pe = cend

            if pe > 0:
                # get build_cardset custom function or default
                custom_build_cardset = "self.%s" % pagination['function']
                CARDS = eval(custom_build_cardset)(pagination['doclist'][ps:pe])
            else:
                CARDS = ""

            if pagination['title'] is None:
                title = pagination['basename'].replace('_', ' ')
            else:
                title = pagination['title']
            content = PG_HEAD % (title, PAGINATION, CARDS)
            if not pagination['fake']:
                self.distribute(name, content)
            pagelist.append(name)
        return pagelist
        # ~ self.log.debug("Created '%s' page (%d pages with %d cards in each page)", pagination['basename'], total_pages, k)

    def build_cardset(self, doclist):
        """Default method to build pages paginated"""
        CARD_DOC_FILTER_DATA_TITLE = self.template('CARD_DOC_FILTER_DATA_TITLE')
        CARDS = ""
        for doc in doclist:
            title = self.srvdtb.get_values(doc, 'Title')[0]
            doc_card = self.get_doc_card(doc)
            card_search_filter = CARD_DOC_FILTER_DATA_TITLE % (valid_filename(title), doc_card)
            CARDS += """%s""" % card_search_filter
        return CARDS

    def create_tagcloud_from_key(self, key):
        """Create a tag cloud based on key values."""
        dkeyurl = {}
        for doc in self.srvdtb.get_documents():
            tags = self.srvdtb.get_values(doc, key)
            url = os.path.basename(doc)[:-5]
            for tag in tags:
                try:
                    urllist = dkeyurl[tag]
                    surllist = set(urllist)
                    surllist.add(url)
                    dkeyurl[tag] = list(surllist)
                except KeyError:
                    surllist = set()
                    surllist.add(url)
                    dkeyurl[tag] = list(surllist)


        max_frequency = set_max_frequency(dkeyurl)
        lwords = []

        for word in dkeyurl:
            len_word = len(word)
            if len_word > 0:
                lwords.append(word)

        len_words = len(lwords)
        if  len_words > 0:
            lwords.sort(key=lambda y: y.lower())

            WORDCLOUD = self.template('WORDCLOUD')
            WORDCLOUD_ITEM = self.template('WORDCLOUD_ITEM')
            items = ''
            for word in lwords:
                frequency = len(dkeyurl[word])
                size = get_font_size(frequency, max_frequency)
                url = "%s_%s.html" % (valid_filename(key), valid_filename(word))
                tooltip = "%d documents" % frequency
                item = WORDCLOUD_ITEM % (url, tooltip, size, word)
                items += item
            html = WORDCLOUD % items
        else:
            html = ''

        return html

    def create_page_index_all(self):
        """Create a page with all documents"""
        doclist = self.srvdtb.get_documents()
        pagination = {}
        pagination['basename'] = 'all'
        pagination['doclist'] = doclist
        pagination['title'] = 'All documents'
        pagination['function'] = 'build_cardset'
        pagination['template'] = 'PAGE_PAGINATION_HEAD'
        pagination['fake'] = False
        self.build_pagination(pagination)

    def create_page_recents(self):
        """Create a page with 60 documents sorted by date desc"""
        doclist = self.srvdtb.get_documents()[:60]
        pagination = {}
        pagination['basename'] = 'recents'
        pagination['doclist'] = doclist
        pagination['title'] = 'Recent documents'
        pagination['function'] = 'build_cardset'
        pagination['template'] = 'PAGE_PAGINATION_HEAD'
        pagination['fake'] = False
        self.build_pagination(pagination)

    def generate_sources(self):
        """Custom themes can use this method to generate source documents"""
        pass

    def create_page_help(self):
        """
        Create help page.
        To be replaced by custom code.
        """
        TPL_HELP = self.template('PAGE_HELP')
        self.distribute('help', TPL_HELP)

    def create_page_index(self):
        """Create index page.
        To be replaced by custom code.
        """
        TPL_INDEX = self.template('PAGE_INDEX')
        self.distribute('index', TPL_INDEX)

    # ~ def create_page_all_keys(self):
        # ~ """Create page as a cloud of keys."""
        # ~ content = self.template('PAGE_KEYS')
        # ~ all_keys = self.srvdtb.get_all_keys()
        # ~ for key in all_keys:
            # ~ cloud = self.create_tagcloud_from_key(key)
            # ~ len_cloud = len(cloud)
            # ~ if  len_cloud > 0:
                # ~ content += "\n\n== %s\n\n" % key
                # ~ content += "\n++++\n%s\n++++\n" % cloud
        # ~ self.distribute('keys', content)

    def get_maxkv_freq(self):
        """Calculate max frequency for all keys"""
        maxkvfreq = 0
        all_keys = self.srvdtb.get_all_keys()
        for key in all_keys:
            blocked_keys = self.srvdtb.get_blocked_keys()
            if key not in blocked_keys:
                values = self.srvdtb.get_all_values_for_key(key)
                if len(values) > maxkvfreq:
                    maxkvfreq = len(values)
        return maxkvfreq

    def create_page_properties(self):
        """Create properties page"""
        TPL_PROPS_PAGE = self.template('PAGE_PROPERTIES')
        TPL_KEY_MODAL_BUTTON = self.template('KEY_MODAL_BUTTON')
        max_frequency = self.get_maxkv_freq()
        all_keys = self.srvdtb.get_all_keys()
        custom_buttons = ''
        for key in all_keys:
            ignored_keys = self.srvdtb.get_ignored_keys()
            if key not in ignored_keys:
                html = self.create_tagcloud_from_key(key)
                values = self.srvdtb.get_all_values_for_key(key)
                frequency = len(values)
                size = get_font_size(frequency, max_frequency)
                proportion = int(math.log((frequency * 100) / max_frequency))
                tooltip = "%d values" % len(values)
                button = TPL_KEY_MODAL_BUTTON % (valid_filename(key), tooltip, size, key, valid_filename(key), valid_filename(key), key, html)
                custom_buttons += button
        content = TPL_PROPS_PAGE % (custom_buttons)
        self.distribute('properties', content)

    def create_page_stats(self):
        """Create stats page"""
        TPL_STATS_PAGE = self.template('PAGE_STATS')
        item = self.template('KEY_LEADER_ITEM')
        numdocs = self.srvapp.get_numdocs()
        keys = self.srvdtb.get_all_keys()
        numkeys = len(keys)
        leader_items = ''
        for key in keys:
            values = self.srvdtb.get_all_values_for_key(key)
            leader_items += item % (valid_filename(key), key, valid_filename(key), len(values))
        stats = TPL_STATS_PAGE % (numdocs, numkeys, leader_items)
        self.distribute('stats', stats)


    def create_page_key(self, key, values):
        """Create key page."""
        source_dir = self.srvapp.get_source_path()
        num_values = len(values)
        html = self.template('PAGE_KEY')

        # TAB Cloud
        cloud = self.create_tagcloud_from_key(key)

        # TAB Stats
        stats = ""
        leader_row = self.template('LEADER_ROW')
        for value in values:
            docs = self.srvdtb.get_docs_by_key_value(key, value)
            tpl_value_link = self.template('LEADER_ROW_VALUE_LINK')
            value_link = tpl_value_link % (valid_filename(key), valid_filename(value), value)
            stats += leader_row % (value_link, len(docs))

        return html % (key, cloud, stats)

    def get_html_values_from_key(self, doc, key):
        """Return the html link for a value."""
        html = []

        values = self.srvdtb.get_values(doc, key)
        for value in values:
            url = "%s_%s.html" % (key, value)
            html.append((url, value))
        return html

    def create_metadata_section(self, doc):
        """Return a html block for displaying metadata (keys and values)."""
        try:
            doc_path = os.path.join(self.srvapp.get_source_path(), doc)
            html = self.template('METADATA_SECTION_HEADER')
            ROW_CUSTOM_PROP = self.template('METADATA_ROW_CUSTOM_PROPERTY')
            ROW_CUSTOM_PROP_TIMESTAMP = self.template('METADATA_ROW_CUSTOM_PROPERTY_TIMESTAMP')
            custom_keys = self.srvdtb.get_custom_keys(doc)
            custom_props = ''
            for key in custom_keys:
                try:
                    values = self.get_html_values_from_key(doc, key)
                    labels = self.get_labels(values)
                    custom_props += ROW_CUSTOM_PROP % (valid_filename(key), key, labels)
                except Exception as error:
                    self.log.error("Key[%s]: %s", key, error)
            timestamp = self.srvdtb.get_doc_timestamp(doc)
            custom_props += ROW_CUSTOM_PROP_TIMESTAMP % ('Timestamp', timestamp)
            num_custom_props = len(custom_props)
            if  num_custom_props > 1:
                html += custom_props
            html += self.template('METADATA_SECTION_FOOTER')
        except Exception as error:
            msgerror = "%s -> %s" % (doc, error)
            self.log.error("%s", msgerror)
            html = ''
            raise

        return html

    def get_doc_card(self, doc):
        """Get card for a given doc"""
        source_dir = self.srvapp.get_source_path()
        DOC_CARD = self.template('CARD_DOC')
        DOC_CARD_FOOTER = self.template('CARD_DOC_FOOTER')
        LINK = self.template('LINK')
        title = self.srvdtb.get_values(doc, 'Title')[0]
        category = self.srvdtb.get_values(doc, 'Category')[0]
        scope = self.srvdtb.get_values(doc, 'Scope')[0]
        link_title = LINK % ("uk-link-heading uk-text-meta", "%s.html" % valid_filename(doc).replace('.adoc', ''), "", title)
        if len(category) > 0 and len(scope) > 0:
            link_category = LINK % ("uk-link-heading uk-text-meta", "Category_%s.html" % valid_filename(category), "", category)
            link_scope = LINK % ("uk-link-heading uk-text-meta", "Scope_%s.html" % valid_filename(scope), "", scope)
            footer = DOC_CARD_FOOTER % (link_category, link_scope)
        else:
            footer = ''

        timestamp = self.srvdtb.get_doc_timestamp(doc)
        fuzzy_date = fuzzy_date_from_timestamp(timestamp)

        tooltip ="%s" % (title)
        return DOC_CARD % (tooltip, link_title, timestamp, fuzzy_date, footer)

    def get_labels(self, values):
        """C0111: Missing function docstring (missing-docstring)."""
        label_links = ''
        value_link = self.template('METADATA_VALUE_LINK')
        for page, text in values:
            if len(text) != 0:
                label_links += value_link % (valid_filename(page), text)
        return label_links

    def create_page_about_app(self):
        """
        About app page.
        To be replaced by custom code.
        """
        page = self.template('PAGE_ABOUT_APP')
        srcdir = self.srvapp.get_source_path()
        about = os.path.join(srcdir, 'about.adoc')
        try:
            content = open(about, 'r').read()
        except:
            content = 'No info available'
        self.distribute('about_app', page % content)

    def create_page_about_theme(self):
        """About theme page."""
        page = self.template('PAGE_ABOUT_THEME')
        theme = self.srvapp.get_theme_properties()
        # ~ content = ''
        # ~ for prop in theme:
            # ~ field_prop = "| *%s*\n" % prop
            # ~ field_value = "| %s\n\n" % theme[prop]
            # ~ content += "%s%s" % (field_prop, field_value)
        self.distribute('about_theme', page % (theme['name'], theme['description'], theme['version']))

    def create_page_about_kb4it(self):
        """About KB4IT page."""
        page = self.template('PAGE_ABOUT_KB4IT')
        self.distribute('about_kb4it', page % APP['version'])




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
KB4IT module. Entry point.

# Author: Tomás Vírseda <tomasvirseda@gmail.com>
# Version: 0.4
# License: GPLv3
# Description: Build a static documentation site based on Asciidoctor
"""

import os
import sys
import argparse
from argparse import Namespace
from kb4it.core.env import APP, LPATH, GPATH
from kb4it.core.log import get_logger
from kb4it.services.app import KB4ITApp
from kb4it.services.database import KB4ITDB
from kb4it.services.builder import KB4ITBuilder

class KB4IT:
    """
    KB4IT main class

    It can be executed from command line:
    $HOME/.local/bin/kb4it -theme <None|THEME> -force -log DEBUG -sort <ATTRIBUTE> -source <SOURCE_PATH> -target <TARGET_PATH>

    Or it can be called from another app as a library:
    from kb4it.kb4it import KB4IT
    from argparse import Namespace
    params = Namespace(FORCE=True, LOGLEVEL='INFO', SORT_ATTRIBUTE=None, SOURCE_PATH='/tmp/myapp', TARGET_PATH='/tmp/output', THEME=None)
    kb = KB4IT(params)
    kb.run()
    """

    ready = False
    params = None
    log = None
    source_path = None
    target_path = None
    numdocs = 0
    tmpdir = None

    def __init__(self, params=None):
        """Initialize KB4IT class."""
        if params is not None:
            self.params = params
        else:
            self.params = Namespace()
        try:
            self.setup_logging(params.LOGLEVEL)
        except:
            self.setup_logging('INFO')
        self.check_params()
        self.setup_services()
        self.setup_environment()

    def check_params(self):
        try:
            source = os.path.abspath(self.params.SOURCE_PATH)
            target = os.path.abspath(self.params.TARGET_PATH)
            if source == target:
                self.log.error("Error. Source and target paths are the same.")
                self.log.error("Source path: %s", source)
                self.log.error("Target path: %s", target)
                self.log.error("Check, please!")
                sys.exit(-1)
            self.ready = True
        except:
            pass

    def get_params(self):
        """Return parametres."""
        return self.params

    def setup_environment(self):
        """Set up KB4IT environment."""
        self.log.debug("Setting up %s environment", APP['shortname'])
        self.log.debug("Global path: %s", GPATH['ROOT'])
        self.log.debug("Local path: %s", LPATH['ROOT'])

        # Create local paths if they do not exist
        for entry in LPATH:
            self.log.debug("Checking if directory '%s' exists", LPATH[entry])
            if not os.path.exists(LPATH[entry]):
                os.makedirs(LPATH[entry])
                self.log.debug("Creating directory '%s'", LPATH[entry])

    def setup_logging(self, severity=None):
        """Set up logging."""
        self.log = get_logger(__class__.__name__, severity.upper())
        self.log.debug("Log level set to: %s", severity.upper())

    def setup_services(self):
        """Set up services."""
        # Declare and register services
        self.services = {}
        try:
            services = {
                'DB': KB4ITDB(),
                'App': KB4ITApp(),
                'Builder': KB4ITBuilder(),
            }
            for name in services:
                self.register_service(name, services[name])
        except Exception as error:
            self.log.error(error)
            raise

    def get_service(self, name):
        """Get or start a registered service."""
        try:
            service = self.services[name]
            logname = service.__class__.__name__
            if not service.is_started():
                service.start(self, logname, name)
            return service
        except KeyError as service:
            self.log.error("Service %s not registered or not found", service)
            raise

    def register_service(self, name, service):
        """Register a new service."""
        try:
            self.services[name] = service
            self.log.debug("Service '%s' registered", name)
        except KeyError as error:
            self.log.error(error)

    def deregister_service(self, name):
        """Deregister a running service."""
        self.services[name].end()
        self.services[name] = None

    def check_parameters(self, params):
        """Check paramaters from command line."""
        self.params = params
        self.source_path = params.SOURCE_PATH
        self.target_path = params.TARGET_PATH
        if self.target_path is None:
            self.target_path = os.path.abspath(os.path.curdir + '/target')
            self.log.debug("\tNo target path provided. Using: %s", self.target_path)

        if not os.path.exists(self.target_path):
            os.makedirs(self.target_path)
            self.log.debug("\tTarget path %s created", self.target_path)

        self.log.debug("\tScript directory: %s", GPATH['ROOT'])
        self.log.debug("\tResources directory: %s", GPATH['RESOURCES'])
        self.log.debug("\tSource directory: %s", self.source_path)
        self.log.debug("\tTarget directory: %s", self.target_path)
        self.log.debug("\tTemporary directory: %s", self.tmpdir)

    def run(self):
        """Start application."""
        if self.ready:
            srvapp = self.get_service('App')
            if self.params.RESET:
                if self.params.FORCE:
                    srvapp.reset()
                else:
                    self.log.warning("KB4IT environment NOT reset. You must force it!")
            else:
                srvapp.run()
                self.stop()

    def get_version(self):
        return '%s %s' % (APP['shortname'], APP['version'])

    def stop(self):
        """Stop registered services by executing the 'end' method (if any)."""
        for name in self.services:
            self.deregister_service(name)
        self.log.debug("KB4IT %s finished", APP['version'])


def main():
    """Execute application."""
    parser = argparse.ArgumentParser(description='KB4IT %s by Tomás Vírseda' % APP['version'])
    parser.add_argument('-reset', action='store_true', dest='RESET', help='Reset environment')
    parser.add_argument('-force', action='store_true', dest='FORCE', help='Force a clean compilation')
    parser.add_argument('-theme', dest='THEME', help='Specify theme. Otherwise, it uses the default one', required=False)
    parser.add_argument('-source', dest='SOURCE_PATH', help='Source directory with asciidoctor source files', required=True)
    parser.add_argument('-target', dest='TARGET_PATH', help='Target directory', required=True)
    parser.add_argument('-sort', dest='SORT_ATTRIBUTE', help='Choose another attribute for sorting instead the default timestamp')
    parser.add_argument('-log', dest='LOGLEVEL', help='Increase output verbosity', action='store', default='INFO')
    parser.add_argument('-version', action='version', version='%s %s' % (APP['shortname'], APP['version']))
    params = parser.parse_args()
    app = KB4IT(params)
    app.run()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module with the application logic.

# Author: Tomás Vírseda <tomasvirseda@gmail.com>
# License: GPLv3
# Description: module holding the application logic
"""

import os
import sys
import glob
import time
import json
import random
import threading
import shutil
import tempfile
import datetime
import operator
from concurrent.futures import ThreadPoolExecutor as Executor

from kb4it.core.env import LPATH, GPATH, APP, ADOCPROPS, MAX_WORKERS, EOHMARK
from kb4it.core.service import Service
from kb4it.core.util import get_human_datetime
from kb4it.core.util import extract_toc, valid_filename, load_kbdict
from kb4it.core.util import exec_cmd, delete_target_contents
from kb4it.core.util import get_source_docs, get_asciidoctor_attributes, get_hash_from_dict
from kb4it.core.util import save_kbdict, copy_docs, copydir
from kb4it.core.util import file_timestamp
from kb4it.core.util import guess_datetime, string_timestamp


class KB4ITApp(Service):
    """C0111: Missing function docstring (missing-docstring)."""
    runtime = {} # Dictionary of runtime properties
    kbdict_new = {} # New compilation cache
    kbdict_cur = {} # Cached data

    def initialize(self):
        """Initialize application structure"""

        # Get params from command line
        self.parameters = self.app.get_params()
        self.log.debug("KB4IT params: %s", self.app.get_params())

        # Initialize directories
        self.runtime['dir'] = {}
        self.runtime['dir']['tmp'] = tempfile.mkdtemp(prefix=LPATH['TMP']+'/')
        if self.parameters.TARGET_PATH is None:
            self.runtime['dir']['target'] = LPATH['WWW']
        else:
            self.runtime['dir']['target'] = os.path.realpath(self.parameters.TARGET_PATH)
        self.runtime['dir']['source'] = os.path.realpath(self.parameters.SOURCE_PATH)
        self.runtime['dir']['cache'] = os.path.join(LPATH['CACHE'], valid_filename(self.runtime['dir']['source']))
        if not os.path.exists(self.runtime['dir']['cache']):
            os.makedirs(self.runtime['dir']['cache'])

        # if SORT attribute is given, use it instead of the OS timestamp
        if self.parameters.SORT_ATTRIBUTE is None:
            self.runtime['sort_attribute'] = 'Timestamp'
        else:
            self.runtime['sort_attribute'] = self.parameters.SORT_ATTRIBUTE
        self.log.debug("Sort attribute: %s", self.runtime['sort_attribute'])

        # Initialize docs structure
        self.runtime['docs'] = {}
        self.runtime['docs']['count'] = 0
        self.runtime['docs']['bag'] = []
        self.runtime['docs']['cached'] = []

        # Load cache dictionary and initialize the new one
        self.kbdict_cur = load_kbdict(self.runtime['dir']['source'])
        self.kbdict_new['document'] = {}
        self.kbdict_new['metadata'] = {}

        # Get services
        self.get_services()

        # Select theme
        # ~ self.theme_load()

    def load_theme(self):
        """Load custom user theme, global theme or default."""

        # custom theme requested by user via command line properties
        self.runtime['theme'] =  {}
        self.runtime['theme']['path'] = self.theme_search(self.parameters.THEME)
        if self.runtime['theme']['path'] is None:
            self.runtime['theme']['path'] = os.path.join(GPATH['THEMES'], 'default')
            self.log.warning("Fallback to default theme")

        theme_conf = os.path.join(self.runtime['theme']['path'], "theme.adoc")
        if not os.path.exists(theme_conf):
            self.log.error("Theme config file not found: %s", theme_conf)
            sys.exit(-1)

        # load theme configuration
        with open(theme_conf, 'r') as fth:
            theme = json.load(fth)
            for prop in theme:
                self.runtime['theme'][prop] = theme[prop]

        self.log.debug("Theme %s (%s) v%s for KB4IT v%s", theme['name'], theme['description'], theme['version'], theme['kb4it'])
        # ~ for prop in theme:
            # ~ if prop != 'name':
                # ~ self.log.debug("%s: %s", prop.title(), theme[prop])

        # Get theme directories
        self.runtime['theme']['templates'] = os.path.join(self.runtime['theme']['path'], 'templates')
        self.runtime['theme']['logic'] = os.path.join(self.runtime['theme']['path'], 'logic')

        # Get date-based attributes from theme. Date attributes aren't
        # displayed as properties but used to build events pages.
        try:
            ignored_keys = self.runtime['theme']['ignored_keys']
            for key in ignored_keys:
                self.log.debug("Ignoring key: %s", key)
                self.srvdtb.ignore_key(key)
        except KeyError:
            self.log.debug("No ignored_keys defined in this theme")

        # Register theme service
        sys.path.insert(0, self.runtime['theme']['logic'])
        try:
            from theme import Theme
            self.app.register_service('Theme', Theme())
            self.srvthm = self.get_service('Theme')
        except Exception as error:
            self.log.warning("Theme scripts for '%s' couldn't be loaded", self.runtime['theme']['id'])
            self.log.error(error)
            raise
        self.log.debug("Loaded theme '%s': %s", self.runtime['theme']['id'], self.runtime['theme']['path'])

    def theme_search(self, theme):
        """Search custom theme"""

        if theme is None:
            # No custom theme passed in arguments. Autodetect.
            self.log.debug("Autodetecting theme from source path")
            source_path = self.runtime['dir']['source']
            source_resources_path = os.path.join(source_path, 'resources')
            source_themes_path = os.path.join(source_resources_path, 'themes')
            all_themes = os.path.join(source_themes_path, '*')
            self.log.debug("Looking for first theme ocurrence in: %s", all_themes)
            try:
                theme_path = glob.glob(all_themes)[0]
                self.log.debug("Found theme path: %s", theme_path)
                return theme_path
            except:
                theme_path = None
                self.log.debug("Theme not found")
                return None

        found = False

        # Search in sources path
        source_path = self.runtime['dir']['source']
        theme_rel_path = os.path.join(os.path.join('resources', 'themes'), theme)
        theme_path = os.path.join(self.runtime['dir']['source'], theme_rel_path)
        if os.path.exists(theme_path):
            found = True
        else:
            # Search for theme in KB4IT global theme
            theme_path = os.path.join(GPATH['THEMES'], theme)
            if os.path.exists(theme_path):
                found = True

        if found:
            # Return custom theme
            self.log.debug("Found theme %s in local repository: %s" % (theme, theme_path))
            return theme_path
        else:
            # Fallback to default
            self.log.debug("Theme not found in local repository: %s." % theme)
            return None

    def get_runtime_properties(self):
        return self.runtime

    def get_runtime_parameter(self, parameter):
        return self.runtime[parameter]

    def get_theme_properties(self):
        return self.runtime['theme']

    def get_theme_property(self, prop):
        return self.runtime['theme'][prop]

    def get_cache_path(self):
        return self.runtime['dir']['cache']

    def get_source_path(self):
        """Get asciidoctor sources path."""
        return self.runtime['dir']['source']

    def get_target_path(self):
        """Get target path."""
        return self.runtime['dir']['target']

    def get_temp_path(self):
        """Get temporary working path."""
        return self.runtime['dir']['tmp']

    def get_services(self):
        """C0111: Missing function docstring (missing-docstring)."""
        self.srvdtb = self.get_service('DB')
        self.srvbld = self.get_service('Builder')

    def get_numdocs(self):
        """C0111: Missing function docstring (missing-docstring)."""
        return self.runtime['docs']['count']

    def highlight_metadata_section(self, source):
        """C0111: Missing function docstring (missing-docstring)."""
        content = source.replace(self.srvbld.template('HTML_TAG_METADATA_OLD'), self.srvbld.template('HTML_TAG_METADATA_NEW'), 1)
        return content

    def apply_transformations(self, source):
        """C0111: Missing function docstring (missing-docstring)."""
        content = source.replace(self.srvbld.template('HTML_TAG_TOC_OLD'), self.srvbld.template('HTML_TAG_TOC_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_SECT1_OLD'), self.srvbld.template('HTML_TAG_SECT1_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_SECT2_OLD'), self.srvbld.template('HTML_TAG_SECT2_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_SECT3_OLD'), self.srvbld.template('HTML_TAG_SECT3_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_SECT4_OLD'), self.srvbld.template('HTML_TAG_SECT4_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_SECTIONBODY_OLD'), self.srvbld.template('HTML_TAG_SECTIONBODY_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_PRE_OLD'), self.srvbld.template('HTML_TAG_PRE_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_H2_OLD'), self.srvbld.template('HTML_TAG_H2_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_H3_OLD'), self.srvbld.template('HTML_TAG_H3_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_H4_OLD'), self.srvbld.template('HTML_TAG_H4_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_TABLE_OLD'), self.srvbld.template('HTML_TAG_TABLE_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_TABLE_OLD_2'), self.srvbld.template('HTML_TAG_TABLE_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_ADMONITION_ICON_NOTE_OLD'), self.srvbld.template('HTML_TAG_ADMONITION_ICON_NOTE_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_ADMONITION_ICON_TIP_OLD'), self.srvbld.template('HTML_TAG_ADMONITION_ICON_TIP_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_ADMONITION_ICON_IMPORTANT_OLD'), self.srvbld.template('HTML_TAG_ADMONITION_ICON_IMPORTANT_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_ADMONITION_ICON_CAUTION_OLD'), self.srvbld.template('HTML_TAG_ADMONITION_ICON_CAUTION_NEW'))
        content = content.replace(self.srvbld.template('HTML_TAG_ADMONITION_ICON_WARNING_OLD'), self.srvbld.template('HTML_TAG_ADMONITION_ICON_WARNING_NEW'))
        return content

    def stage_01_check_environment(self):
        """Check environment."""
        self.log.debug("Stage 1\tCheck environment")
        self.log.debug("Cache directory: %s", self.runtime['dir']['cache'])
        self.log.debug("Working directory: %s", self.runtime['dir']['tmp'])
        self.log.debug("Source directory: %s", self.runtime['dir']['source'])
        try:
            self.runtime['theme']
        except:
            self.load_theme()

        # check if target directory exists. If not, create it:
        if not os.path.exists(self.runtime['dir']['target']):
            os.makedirs(self.runtime['dir']['target'])
        self.log.debug("Target directory: %s", self.runtime['dir']['target'])

    def stage_02_get_source_documents(self):
        """Get Asciidoctor source documents."""
        self.log.debug("Stage 2\tGet Asciidoctor source documents")
        self.runtime['docs']['bag'] = get_source_docs(self.runtime['dir']['source'])
        self.runtime['docs']['count'] = len(self.runtime['docs']['bag'])
        self.log.debug("Found %d asciidoctor documents", self.runtime['docs']['count'])
        for doc in self.runtime['docs']['bag']:
            self.log.debug("%s", doc)

    def stage_03_preprocessing(self):
        """
        Extract metadata from source docs into a dict.

        Create metadata section for each adoc and insert it after the
        EOHMARK.

        In this way, after being compiled into HTML, final adocs are
        browsable throught its metadata.
        """
        self.log.debug("Stage 3\tPreprocessing")
        self.log.debug("Clean up cache")

        def clean_cache():
            missing = []
            try:
                for docname in self.kbdict_cur['document']:
                    docpath = os.path.join(self.runtime['dir']['source'], docname)
                    if not os.path.exists(docpath):
                        missing.append(docname)
            except:
                self.log.debug("Cache is empty")

            for docname in missing:
                docname = docname.replace('.adoc', '')
                self.delete_document(docname)

        clean_cache()

        tsdict = {}
        for source in self.runtime['docs']['bag']:
            docname = os.path.basename(source)
            self.kbdict_new['document'][docname] = {}
            self.log.debug("? DOC[%s] Preprocessing", docname)

            # Add a new document to the database
            self.srvdtb.add_document(docname)

            # Get datetime timestamp from filesystem and add it as
            # attribute
            timestamp = file_timestamp(source)
            self.srvdtb.add_document_key(docname, 'Timestamp', timestamp)

            # Get content
            with open(source) as source_adoc:
                srcadoc = source_adoc.read()

            # Get metadata
            docpath = os.path.join(self.runtime['dir']['source'], docname)
            keys = get_asciidoctor_attributes(docpath)

            # To track changes in a document, hashes for metadata and content are created.
            # Comparing them with those in the cache, KB4IT determines if a document must be
            # compiled again. Very useful to reduce the compilation time.

            # Get Document Content and Metadata Hashes
            self.kbdict_new['document'][docname]['content_hash'] = get_hash_from_dict({'content': srcadoc})
            self.kbdict_new['document'][docname]['metadata_hash'] = get_hash_from_dict(keys)
            self.kbdict_new['document'][docname]['Timestamp'] = timestamp

            # Generate caches
            for key in keys:
                alist = keys[key]
                for value in alist:
                    nc = len(value.strip())
                    if nc == 0:
                        continue
                    try:
                        if key in self.runtime['theme']['date_attributes']:
                            value = string_timestamp(value)
                    except:
                        pass

                    if key == 'Tag':
                        value = value.lower()
                    self.srvdtb.add_document_key(docname, key, value)

                    # For each document and for each key/value linked to that document add an entry to kbdic['document']
                    try:
                        values = self.kbdict_new['document'][docname][key]
                        if value not in values:
                            values.append(value)
                        self.kbdict_new['document'][docname][key] = sorted(values)
                    except:
                        self.kbdict_new['document'][docname][key] = [value]

                    # And viceversa, for each key/value add to kbdict['metadata'] all documents linked
                    try:
                        documents = self.kbdict_new['metadata'][key][value]
                        documents.append(docname)
                        self.kbdict_new[key][value] = sorted(documents, key=lambda y: y.lower())
                    except KeyError:
                        if key not in self.kbdict_new['metadata']:
                            self.kbdict_new['metadata'][key] = {}
                        if value not in self.kbdict_new['metadata'][key]:
                            self.kbdict_new['metadata'][key][value] = [docname]

            # Force compilation (from command line)?
            DOC_COMPILATION = False
            FORCE_ALL = self.parameters.FORCE
            if not FORCE_ALL:
                # Get cached document path and check if it exists
                cached_document = os.path.join(self.runtime['dir']['cache'], docname.replace('.adoc', '.html'))
                cached_document_exists = os.path.exists(cached_document)
                # ~ self.log.debug("* DOC[%s] Cached? %s", docname, cached_document_exists)

                # Compare the document with the one in the cache
                # ~ self.log.debug("Does document %s exist in cache? %s", docname, cached_document_exists)

                if not cached_document_exists:
                    DOC_COMPILATION = True
                    REASON = "Not cached"
                    # ~ self.log.debug("DOC[%s] not cached. Compile", docname)
                else:
                    try:
                        hash_new = self.kbdict_new['document'][docname]['content_hash'] + self.kbdict_new['document'][docname]['metadata_hash']
                        hash_cur = self.kbdict_cur['document'][docname]['content_hash'] + self.kbdict_cur['document'][docname]['metadata_hash']
                        DOC_COMPILATION = hash_new != hash_cur
                        REASON = "Hashes differ? %s" % DOC_COMPILATION
                        # ~ self.log.debug("DOC[%s] cached. Hashes differ? %s. Compile? %s", docname, DOC_COMPILATION, DOC_COMPILATION)
                    except Exception as warning:
                        DOC_COMPILATION = True
                        REASON = warning
            else:
                REASON = "Forced"
                # ~ self.log.debug("* DOC[%s] Compile? %s - Reason: %s", docname, DOC_COMPILATION, REASON)

            COMPILE = DOC_COMPILATION or FORCE_ALL
            # Save compilation status
            self.kbdict_new['document'][docname]['compile'] = COMPILE

            if COMPILE:
                newadoc = srcadoc.replace(EOHMARK, '', 1)

                # Write new adoc to temporary dir
                target = "%s/%s" % (self.runtime['dir']['tmp'], valid_filename(docname))
                self.log.debug("+ DOC[%s] Compile? %s. Reason: %s", docname, COMPILE, REASON)
                with open(target, 'w') as target_adoc:
                    target_adoc.write(newadoc)
            else:
                filename = os.path.join(self.runtime['dir']['cache'], docname.replace('.adoc', '.html'))
                self.runtime['docs']['cached'].append(filename)
                self.log.debug("= DOC[%s] Compile? %s. Reason: %s", docname, COMPILE, REASON)

        # Save current status for the next run
        save_kbdict(self.kbdict_new, self.runtime['dir']['source'])

        # Build a list of documents sorted by timestamp
        self.srvdtb.sort_database()

         # Documents preprocessing stats
        self.log.debug("[PRE-PROCESSING STATS] Number of documents analyzed: %d", len(self.runtime['docs']['bag']))
        keep_docs = compile_docs = 0
        for docname in self.kbdict_new['document']:
            if self.kbdict_new['document'][docname]['compile']:
                compile_docs += 1
            else:
                keep_docs += 1
        self.log.debug("[PRE-PROCESSING STATS] Keep: %d - Compile: %d", keep_docs, compile_docs)

    def get_kbdict_value(self, key, value, new=True):
        if new:
            kbdict = self.kbdict_new
        else:
            kbdict = self.kbdict_cur

        try:
            alist = kbdict['metadata'][key][value]
        except:
            alist = []
        return alist

    def stage_04_processing(self):
        """Process all documents."""
        self.log.debug("Stage 4\tProcessing keys")
        all_keys = set(self.srvdtb.get_all_keys())
        ign_keys = set(self.srvdtb.get_ignored_keys())
        available_keys = list(all_keys - ign_keys)
        K_PATH = []
        KV_PATH = []

        for key in sorted(available_keys):
            COMPILE_KEY = False
            FORCE_ALL = self.parameters.FORCE
            values = self.srvdtb.get_all_values_for_key(key)
            for value in values:
                COMPILE_VALUE = False
                key_value_docs_new = self.get_kbdict_value(key, value, new=True)
                key_value_docs_cur = self.get_kbdict_value(key, value, new=False)
                if key_value_docs_new != key_value_docs_cur:
                    COMPILE_VALUE = True
                COMPILE_VALUE = COMPILE_VALUE or FORCE_ALL
                COMPILE_KEY = COMPILE_KEY or COMPILE_VALUE
                KV_PATH.append((key, value, COMPILE_VALUE))
                self.log.debug("* KEY[%s] VALUE[%s] Compile? %s", key, value, COMPILE_VALUE)
            COMPILE_KEY = COMPILE_KEY or FORCE_ALL
            K_PATH.append((key, values, COMPILE_KEY))
            self.log.debug("* KEY[%s] Compile? %s", key, COMPILE_KEY)

        # To compile or not to compile :)
        ## Keys
        for kpath in K_PATH:
            key, values, COMPILE_KEY = kpath
            if COMPILE_KEY:
                docname = "%s/%s.adoc" % (self.runtime['dir']['tmp'], valid_filename(key))
                html = self.srvthm.create_page_key(key, values)
                with open(docname, 'w') as fkey:
                    fkey.write(html)
            else:
                # Add key to cache
                docname = "%s.html" % (valid_filename(key))
                filename = os.path.join(self.runtime['dir']['cache'], docname)
                self.runtime['docs']['cached'].append(filename)

        ## Keys/Values
        for kvpath in KV_PATH:
            key, value, COMPILE_VALUE = kvpath
            docs = self.get_kbdict_value(key, value, new=True)
            sorted_docs = self.srvdtb.sort_by_date(docs)
            basename = "%s_%s" % (valid_filename(key), valid_filename(value))
            pagination = {}
            pagination['basename'] = basename
            pagination['doclist'] = sorted_docs
            pagination['title'] = None
            pagination['function'] = 'build_cardset'
            pagination['template'] = 'PAGE_PAGINATION_HEAD'
            if COMPILE_VALUE:
                pagination['fake'] = False
                self.srvthm.build_pagination(pagination)
            else:
                pagination['fake'] = True
                pagelist = self.srvthm.build_pagination(pagination)
                for page in pagelist:
                    docname = "%s.html" % page
                    filename = os.path.join(self.runtime['dir']['cache'], docname)
                    self.runtime['docs']['cached'].append(filename)

        self.srvthm.build()

    def stage_05_compilation(self):
        """Compile documents to html with asciidoctor."""
        self.log.debug("Stage 5\tCompilation")
        dcomps = datetime.datetime.now()

        # copy online resources to target path
        # ~ resources_dir_source = GPATH['THEMES']
        resources_dir_tmp = os.path.join(self.runtime['dir']['tmp'], 'resources')
        shutil.copytree(GPATH['RESOURCES'], resources_dir_tmp)
        self.log.debug("Resources copied to '%s'", resources_dir_tmp)

        adocprops = ''
        for prop in ADOCPROPS:
            if ADOCPROPS[prop] is not None:
                if '%s' in ADOCPROPS[prop]:
                    adocprops += '-a %s=%s ' % (prop, ADOCPROPS[prop] % self.runtime['dir']['target'])
                else:
                    adocprops += '-a %s=%s ' % (prop, ADOCPROPS[prop])
            else:
                adocprops += '-a %s ' % prop
        self.log.debug("Parameters passed to Asciidoctor: %s", adocprops)

        with Executor(max_workers=MAX_WORKERS) as exe:
            docs = get_source_docs(self.runtime['dir']['tmp'])
            jobs = []
            jobcount = 0
            num = 1
            self.log.debug("Generating jobs. Please, wait")
            for doc in docs:
                cmd = "asciidoctor -q -s %s -b html5 -D %s %s" % (adocprops, self.runtime['dir']['tmp'], doc)
                job = exe.submit(exec_cmd, (doc, cmd, num))
                job.add_done_callback(self.srvthm.build_page)
                self.log.debug("Job[%4d]: %s will be compiled", num, os.path.basename(doc))
                # ~ self.log.debug("Job[%4d]: %s", num, cmd)
                jobs.append(job)
                num = num + 1
            self.log.debug("Created %d jobs. Starting compilation", num - 1)
            # ~ self.log.debug("%3s%% done", "0")
            for job in jobs:
                adoc, res, jobid = job.result()
                self.log.debug("Job[%d/%d]: %s compiled successfully", jobid, num - 1, os.path.basename(adoc))
                jobcount += 1
                if jobcount % MAX_WORKERS == 0:
                    pct = int(jobcount * 100 / len(docs))
                    self.log.debug("%3s%% done", str(pct))

        dcompe = datetime.datetime.now()
        totaldocs = len(get_source_docs(self.runtime['dir']['tmp']))
        comptime = dcompe - dcomps
        self.log.debug("100% done")
        self.log.debug("Compilation time: %d seconds", comptime.seconds)
        self.log.debug("Number of compiled docs: %d", totaldocs)
        try:
            self.log.debug("Compilation Avg. Speed: %d docs/sec",
                          int((totaldocs/comptime.seconds)))
        except ZeroDivisionError:
            self.log.debug("Compilation Avg. Speed: %d docs/sec",
                          int((totaldocs/1)))

    def stage_06_extras(self):
        """Include other stuff."""
        pass

    def stage_07_clean_target(self):
        """Delete contents of target directory (if any)."""
        self.log.debug("Stage 6\tClean target directory")
        delete_target_contents(self.runtime['dir']['target'])
        self.log.debug("Deleted target contents in: %s", self.runtime['dir']['target'])

    def stage_08_refresh_target(self):
        """Refresh target directory."""
        self.log.debug("Stage 7\tRefresh target directory")

        # Copy asciidocs documents to target path
        pattern = os.path.join(self.runtime['dir']['source'], '*.adoc')
        files = glob.glob(pattern)
        docsdir = os.path.join(self.runtime['dir']['target'], 'sources')
        os.makedirs(docsdir)
        copy_docs(files, docsdir)
        self.log.debug("Copy %d asciidoctor sources from source path to target path", len(files))

        # Copy compiled documents to target path
        pattern = os.path.join(self.runtime['dir']['tmp'], '*.html')
        files = glob.glob(pattern)
        copy_docs(files, self.runtime['dir']['target'])
        self.log.debug("Copy %d html files from temporary path to target path", len(files))

        # Copy cached documents to target path
        copy_docs(self.runtime['docs']['cached'], self.runtime['dir']['target'])
        self.log.debug("Copied %d cached documents successfully to target path", len(self.runtime['docs']['cached']))

        # ???
        # ~ copy_docs(self.runtime['docs']['bag'], self.runtime['dir']['target'])
        # ~ self.log.debug("Copied %d Asciidoctor source docs copied to target path", len(self.runtime['docs']['bag']))

        # Copy global resources to target path
        # FIXME: copy common resources, default theme and choosen theme
        resources_dir_target = os.path.join(self.runtime['dir']['target'], 'resources')
        global_resources_dir = GPATH['RESOURCES']
        # ~ COMMON_RES_DIR = os.path.join(global_resources_dir, 'common')
        theme_target_dir = os.path.join(resources_dir_target, 'themes')
        theme = self.get_theme_properties()
        DEFAULT_THEME = os.path.join(GPATH['THEMES'], 'default')
        CUSTOM_THEME_ID = theme['id']
        CUSTOM_THEME_PATH = theme['path']
        copydir(DEFAULT_THEME, os.path.join(theme_target_dir, 'default'))
        copydir(CUSTOM_THEME_PATH, os.path.join(theme_target_dir, CUSTOM_THEME_ID))
        copydir(GPATH['COMMON'], os.path.join(resources_dir_target, 'common'))
        self.log.debug("Copied global resources to target path")

        # Copy local resources to target path
        source_resources_dir = os.path.join(self.runtime['dir']['source'], 'resources')
        if os.path.exists(source_resources_dir):
            resources_dir_target = os.path.join(self.runtime['dir']['target'], 'resources')
            copydir(source_resources_dir, resources_dir_target)
            self.log.debug("Copied local resources to target path")

        # Copy back all HTML files from target to cache
        # Fixme: should cache contents be deleted before copying?
        pattern = os.path.join(self.runtime['dir']['target'], '*.html')
        html_files = glob.glob(pattern)
        copy_docs(html_files, self.runtime['dir']['cache'])
        self.log.debug("Copying HTML files to cache...")

        # Copy JSON database to target path so it can be queried from others applications
        save_kbdict(self.kbdict_new, self.runtime['dir']['target'], 'kb4it')
        self.log.debug("Copied JSON database to target")

    def stage_09_remove_temporary_dir(self):
        """Remove temporary dir."""
        self.log.debug("Stage 8\tRemove temporary directory")
        shutil.rmtree(self.runtime['dir']['tmp'])
        self.log.debug("Temporary directory %s deleted successfully", self.runtime['dir']['tmp'])

    def reset(self):
        """
        WARNING!!!
        Reset environment given source and target directories.
        Delete:
        - Source directory
        - Target directory
        - Temporary directory
        - Cache directory
        - KB4IT database file for this environment
        WARNING!!!

        Please, note: if you pass the wrong directory...
        """
        KB4IT_DB_FILE = os.path.join(LPATH['DB'], 'kbdict-%s.json' % valid_filename(self.runtime['dir']['source']))
        delete_target_contents(self.runtime['dir']['cache'])
        self.kbdict_new = {}
        self.kbdict_cur = {}
        self.log.info("DIR[%s] deleted", self.runtime['dir']['cache'])
        delete_target_contents(self.runtime['dir']['tmp'])
        self.log.info("DIR[%s] deleted", self.runtime['dir']['tmp'])
        delete_target_contents(self.runtime['dir']['source'])
        self.log.info("DIR[%s] deleted", self.runtime['dir']['source'])
        delete_target_contents(self.runtime['dir']['target'])
        self.log.info("DIR[%s] deleted", self.runtime['dir']['target'])
        delete_target_contents(KB4IT_DB_FILE)
        self.log.info("FILE[%s] deleted", KB4IT_DB_FILE)
        self.log.info("KB4IT environment reset")


    def run(self):
        """Start script execution following this flow.

        1. Check environment
        2. Get source documents
        3. Preprocess documents (get metadata)
        4. Process documents in a temporary dir
        5. Compile documents to html with asciidoctor
        6. Delete contents of target directory (if any)
        7. Refresh target directory
        8. Remove temporary directory
        """
        self.log.debug("KB4IT - Knowledge Base for IT")

        self.running = True
        self.log.info("KB4IT - Execution started")
        self.stage_01_check_environment()
        self.srvthm.generate_sources()
        self.stage_02_get_source_documents()
        self.stage_03_preprocessing()
        self.stage_04_processing()
        self.stage_05_compilation()
        self.stage_06_extras()
        self.stage_07_clean_target()
        self.stage_08_refresh_target()
        self.stage_09_remove_temporary_dir()
        self.log.info("KB4IT - Execution finished")
        self.log.debug("Browse your documentation repository:")
        self.log.debug("sensible-browser %s/index.html", os.path.abspath(self.runtime['dir']['target']))
        self.running = False

    def is_running(self):
        return self.running

    def delete_document(self, adoc):
        # Remove source document
        try:
            source_dir = self.get_source_path()
            source_path = os.path.join(source_dir, "%s.adoc" % adoc)
            os.unlink(source_path)
            self.log.debug("DOC[%s] deleted from source directory", adoc)
        except:
            self.log.debug("DOC[%s] not found in source directory", adoc)

        # Remove database document
        self.srvdtb.del_document(adoc)
        # ~ self.log.debug("DOC[%s] deleted from database", adoc)

        # Remove cache document
        cache_dir = self.get_cache_path()
        cached_path = os.path.join(cache_dir, "%s.html" % adoc)
        try:
            os.unlink(cached_path)
            self.log.debug("DOC[%s] deleted from cache directory", adoc)
        except:
            self.log.debug("DOC[%s] not found in cache directory", adoc)




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Environment module.

# File: mod_env.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: Environment variables module
"""

import os
from os.path import abspath
import sys
import glob


ROOT = abspath(sys.modules[__name__].__file__ + "/../../")
USER_DIR = os.path.expanduser('~')

# App Info
APP = {}
APP['name'] = "Knowledge Base For IT"
APP['shortname'] = "KB4IT"
APP['description'] = "KB4IT is a static website generator based on \
                      Asciidoctor sources mainly for technical \
                      documentation purposes."
APP['license'] = 'GPL v3'
APP['license_long'] = "The code is licensed under the terms of the  GPL v3\n\
                  so you're free to grab, extend, improve and fork the \
                  code\nas you want"
APP['copyright'] = "Copyright \xa9 2019 Tomás Vírseda"
APP['desc'] = ""
APP['version'] = '0.7.6.5.1'
APP['author'] = 'Tomás Vírseda'
APP['author_email'] = 'tomasvirseda@gmail.com'
APP['documenters'] = ["Tomás Vírseda <tomasvirseda@gmail.com>"]
APP['website'] = 'https://github.com/t00m/KB4IT'

# Local paths
LPATH = {}
LPATH['ROOT'] = os.path.join(USER_DIR, ".%s" % APP['shortname'].lower())
LPATH['ETC'] = os.path.join(LPATH['ROOT'], 'etc')
LPATH['VAR'] = os.path.join(LPATH['ROOT'], 'var')
LPATH['DB'] = os.path.join(LPATH['VAR'], 'db')
LPATH['PLUGINS'] = os.path.join(LPATH['VAR'], 'plugins')
LPATH['LOG'] = os.path.join(LPATH['VAR'], 'log')
LPATH['TMP'] = os.path.join(LPATH['VAR'], 'tmp')
LPATH['CACHE'] = os.path.join(LPATH['VAR'], 'cache')
LPATH['DB'] = os.path.join(LPATH['VAR'], 'db')
LPATH['WWW'] = os.path.join(LPATH['VAR'], 'www')
LPATH['EXPORT'] = os.path.join(LPATH['VAR'], 'export')
LPATH['OPT'] = os.path.join(LPATH['ROOT'], 'opt')
LPATH['THEMES'] = os.path.join(LPATH['OPT'], 'themes')

# Global paths
GPATH = {}
GPATH['ROOT'] = ROOT
GPATH['DATA'] = os.path.join(GPATH['ROOT'], 'kb4it')
GPATH['RESOURCES'] = os.path.join(GPATH['ROOT'], 'resources')
GPATH['ONLINE'] = os.path.join(GPATH['RESOURCES'], 'online')
GPATH['IMAGES'] = os.path.join(GPATH['ONLINE'], 'images')
GPATH['COMMON'] = os.path.join(GPATH['RESOURCES'], 'common')
GPATH['THEMES'] = os.path.join(GPATH['RESOURCES'], 'themes')
GPATH['APPDATA'] = os.path.join(GPATH['COMMON'], 'appdata')
GPATH['RES'] = os.path.join(GPATH['DATA'], 'res')


# Configuration, SAP Notes Database and Log files
FILE = {}
FILE['CNF'] = os.path.join(LPATH['ETC'], APP['shortname'].lower(), '.ini')
FILE['LOG'] = os.path.join(LPATH['LOG'] + APP['shortname'].lower(), '.log')

# Compilation stuff
MAX_WORKERS = 30
EOHMARK = """// END-OF-HEADER. DO NOT MODIFY OR DELETE THIS LINE"""
ADOCPROPS = {
    'source-highlighter': 'coderay',
    'toc': 'left',
    'toclevels': '3',
    'icons': 'font',
    'linkcss': None,
}



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Log module.

# File: mod_log.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: log module
"""

import logging


def get_logger(name, level=None):
    """Return a new logger."""
    if level is not None:
        if level == 'DEBUG':
            severity = logging.DEBUG
        elif level == 'INFO':
            severity = logging.INFO
        elif level == 'WARNING':
            severity = logging.WARNING
        elif level == 'ERROR':
            severity = logging.ERROR
        else:
            severity = logging.DEBUG
    else:
        severity = logging.INFO
    pattern = "%(levelname)7s | %(lineno)4d | %(name)-15s | %(asctime)s | %(message)s"
    logging.basicConfig(level=logging.DEBUG, format=pattern)
    log = logging.getLogger(name)
    log.setLevel(severity)

    # ~ log.debug("Logger '%s' started with severity '%s'", name, level)
    # ~ formatter = logging.Formatter(pattern)
    # ~ fhl = logging.FileHandler(FILE['LOG'])
    # ~ fhl.setFormatter(formatter)
    # ~ fhl.setLevel(logging.DEBUG)
    # ~ log.addHandler(fhl)

    return log



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Server module.

# File: theme.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: techdoc theme scripts
"""

import calendar
from calendar import HTMLCalendar
from datetime import datetime

from kb4it.services.builder import KB4ITBuilder
from kb4it.core.util import get_human_datetime, fuzzy_date_from_timestamp
from kb4it.core.util import valid_filename, get_human_datetime, guess_datetime
from evcal import EventsCalendar


class Theme(KB4ITBuilder):
    # ~ def initialize(self):
        # ~ self.log.debug("Hi, Ich bin Thema 'techdoc'")

    def generate_sources(self):
        self.log.warning("Oikos shouldn't call this method...")

    def build(self):
        self.create_page_about_app()
        self.create_page_about_theme()
        self.create_page_about_kb4it()
        self.create_page_help()
        self.create_page_properties()
        self.create_page_stats()
        self.create_page_index_all()
        self.create_page_index()
        self.create_page_bookmarks()
        self.create_page_authors()
        self.app.register_service('EvCal', EventsCalendar())
        self.srvcal = self.get_service('EvCal')
        self.create_page_events()
        # ~ self.create_page_blog()
        self.create_page_recents()

    def get_doc_card_event(self, doc):
        source_dir = self.srvapp.get_source_path()
        DOC_CARD = self.template('CARD_DOC_EVENT')
        DOC_CARD_FOOTER = self.template('CARD_DOC_FOOTER')
        LINK = self.template('LINK')
        title = self.srvdtb.get_values(doc, 'Title')[0]
        category = self.srvdtb.get_values(doc, 'Category')[0]
        scope = self.srvdtb.get_values(doc, 'Scope')[0]
        link_title = DOC_CARD_LINK % (valid_filename(doc).replace('.adoc', ''), title)
        if len(category) > 0 and len(scope) >0:
            link_category = LINK % ("uk-link-heading uk-text-meta", "Category_%s.html" % valid_filename(category), "", category)
            link_scope = LINK % ("uk-link-heading uk-text-meta", "Scope_%s.html" % valid_filename(scope), "", scope)
            footer = DOC_CARD_FOOTER % (link_category, link_scope)
        else:
            footer = ''

        timestamp = self.srvdtb.get_doc_timestamp(doc)
        fuzzy_date = fuzzy_date_from_timestamp(timestamp)
        tooltip ="%s" % (title)
        return DOC_CARD % (tooltip, link_title, timestamp, fuzzy_date, footer)

    def build_events(self, doclist):
        SORT = self.srvapp.get_runtime_parameter('sort_attribute')
        # ~ self.log.debug("\t\t\t  Using custom build cardset function for events")
        dey = {} # Dictionary of day events per year
        events_docs = {} # Dictionary storing a list of docs for a given date
        events_docs_html = {}

        # Get events dates
        for doc in doclist:
            props = self.srvdtb.get_doc_properties(doc)
            timestamp = self.srvdtb.get_doc_timestamp(doc)
            # Build dict of events for a given date as a list of tuples
            # (month, day) indexed by year
            # Also, build a dict to store those docs ocurring in that date
            try:
                timestamp = guess_datetime(timestamp)
                y = timestamp.year
                m = timestamp.month
                d = timestamp.day
                try:
                    days_events = dey[y]
                    days_events.append((m, d))
                except:
                    days_events = []
                    days_events.append((m, d))
                    dey[y] = days_events

                # Build dict of documents
                if not y in events_docs:
                    events_docs[y] = {}
                    events_docs_html[y] = {}

                if not m in events_docs[y]:
                    events_docs[y][m] = {}
                    events_docs_html[y][m] = {}

                if not d in events_docs[y][m]:
                    events_docs[y][m][d] = []
                    events_docs_html[y][m][d] = ''

                docs = events_docs[y][m][d]
                docs.append(doc)
                events_docs[y][m][d] = docs
            except Exception as error:
                # Doc doesn't have a valid date field. Skip it.
                self.log.error(error)
                raise
                # ~ pass

        # Build day event pages
        for year in events_docs:
            for month in events_docs[year]:
                for day in events_docs[year][month]:
                    docs = events_docs[year][month][day]
                    edt = guess_datetime("%4d.%02d.%02d" % (year, month, day))
                    title = edt.strftime("Events on %A, %B %d %Y")
                    EVENT_PAGE_DAY = "events_%4d%02d%02d" % (year, month, day)

                    # create html page
                    pagination = {}
                    pagination['basename'] = EVENT_PAGE_DAY
                    pagination['doclist'] = docs
                    pagination['title'] = title
                    pagination['function'] = 'build_cardset'
                    pagination['template'] = 'PAGE_PAGINATION_HEAD'
                    pagination['fake'] = False
                    self.build_pagination(pagination)

                    # Generate HTML to display into the modal window
                    # ~ events_docs_html[y][m][d] = self.build_html_events(docs)

        # Build month event pages
        for year in events_docs:
            for month in events_docs[year]:
                thismonth = []
                edt = guess_datetime("%4d.%02d.01" % (year, month))
                title = edt.strftime("Events on %B, %Y")
                EVENT_PAGE_MONTH = "events_%4d%02d" % (year, month)
                for day in events_docs[year][month]:
                    thismonth.extend(events_docs[year][month][day])

                # create html page
                pagination = {}
                pagination['basename'] = EVENT_PAGE_MONTH
                pagination['doclist'] = thismonth
                pagination['title'] = title
                pagination['function'] = 'build_cardset'
                pagination['template'] = 'PAGE_PAGINATION_HEAD'
                pagination['fake'] = False
                self.build_pagination(pagination)

        for year in sorted(dey.keys(), reverse=True):
            HTML = self.srvcal.build_year_pagination(dey.keys())
            edt = guess_datetime("%4d.01.01" % year)
            title = edt.strftime("Events on %Y")
            PAGE = self.template('PAGE_EVENTS_YEAR')
            EVENT_PAGE_YEAR = "events_%4d" % year
            self.srvcal.set_events_days(dey[year])
            self.srvcal.set_events_docs(events_docs[year])
            self.srvcal.set_events_html(events_docs_html[year])
            HTML += self.srvcal.formatyearpage(year, 4)
            self.distribute(EVENT_PAGE_YEAR, PAGE % (title, HTML))

        return dey

    def load_events_days(self, events_days, year):
        events_set = set()
        for event_day in events_days:
            events_set.add(event_day)

        for month, day in events_set:
            adate = guess_datetime("%d.%02d.%02d" % (year, month, day))
            # ~ self.log.debug(adate)

    def create_page_events(self):
        self.log.debug("\t\tBuilding events")
        doclist = []
        ecats = {}
        theme = self.srvapp.get_theme_properties()
        try:
            event_types = theme['events']
        except:
            event_types = []
        self.log.info("\t\tEvent types registered for this theme: %s", ', '.join(event_types))
        for doc in self.srvdtb.get_documents():
            category = self.srvdtb.get_values(doc, 'Category')[0]
            if category in event_types:
                try:
                    docs = ecats[category]
                    docs.add(doc)
                    ecats[category] = docs
                except:
                    docs = set()
                    docs.add(doc)
                    ecats[category] = docs

                doclist.append(doc)
                title = self.srvdtb.get_values(doc, 'Title')[0]
        dey = self.build_events(doclist)
        HTML = self.srvcal.build_year_pagination(dey.keys())
        page = self.template('PAGE_EVENTS')
        self.distribute('events', page % HTML)

    def create_page_recents(self):
        """Create recents page."""
        doclist = self.srvdtb.get_documents()[:60]
        pagination = {}
        pagination['basename'] = 'recents'
        pagination['doclist'] = doclist
        pagination['title'] = 'Recents'
        pagination['function'] = 'build_cardset'
        pagination['template'] = 'PAGE_PAGINATION_HEAD'
        pagination['fake'] = False
        self.build_pagination(pagination)

    def create_page_bookmarks(self):
        """Create bookmarks page."""
        doclist = []
        for doc in self.srvdtb.get_documents():
            bookmark = self.srvdtb.get_values(doc, 'Bookmark')[0]
            if bookmark == 'Yes' or bookmark == 'True':
                doclist.append(doc)
        pagination = {}
        pagination['basename'] = 'bookmarks'
        pagination['doclist'] = doclist
        pagination['title'] = 'Bookmarks'
        pagination['function'] = 'build_cardset'
        pagination['template'] = 'PAGE_PAGINATION_HEAD'
        pagination['fake'] = False
        self.build_pagination(pagination)

    def create_page_authors(self):
        PAGE_AUTHOR = self.template('PAGE_AUTHOR')
        SECTION_ETYPE = self.template('PAGE_AUTHOR_SECTION_EVENT_TYPE')
        SWITCHER_ETYPE = self.template('PAGE_AUTHOR_SWITCHER_EVENT_TYPE')
        authors = self.srvdtb.get_all_values_for_key('Author')

        def tab_header(docs):
            author_etypes = set()
            event_types = self.srvapp.get_theme_property('events')
            for doc in docs:
                categories = self.srvdtb.get_values(doc, 'Category')
                for category in categories:
                    author_etypes.add(category)
            header = """<ul class="uk-flex-center" uk-tab>\n"""
            for etype in sorted(list(author_etypes)):
                header += """<li><a href="#">%s</a></li>\n""" % etype.title()
            # ~ header += """<li><a href="#">Others</a></li>\n"""
            header += """</ul>\n"""
            return sorted(list(author_etypes)), header

        for author in authors:
            docs = self.srvdtb.get_docs_by_key_value('Author', author)
            author_etypes, header = tab_header(docs)
            # ~ self.log.error ("%s -> %s", author, author_etypes)

            # Registered event types
            content_author = ''
            for etype in author_etypes:
                items = ''
                sect_items = 0
                for doc in docs:
                    category = self.srvdtb.get_values(doc, 'Category')
                    if etype in category:
                        items += self.get_doc_card_author(doc)
                        sect_items += 1
                content_author += SECTION_ETYPE % (sect_items, len(docs), items)

            # Others categories
            # ~ others = set(docs) - used
            # ~ items = ''
            # ~ for doc in others:
                # ~ items += self.get_doc_card_author(doc)
            # ~ content_author += SECTION_ETYPE % (len(others), items)
            content = SWITCHER_ETYPE % content_author
            PAGE = PAGE_AUTHOR % (author, header, content)
            # ~ self.log.error(content)
            self.distribute(valid_filename("Author_%s" % author), PAGE)


    def get_doc_card_author(self, doc):
        source_dir = self.srvapp.get_source_path()
        DOC_CARD = self.template('CARD_DOC_AUTHOR')
        DOC_CARD_FOOTER = self.template('CARD_DOC_AUTHOR_FOOTER')
        LINK = self.template('LINK')
        title = self.srvdtb.get_values(doc, 'Title')[0]
        category = self.srvdtb.get_values(doc, 'Category')[0]
        scope = self.srvdtb.get_values(doc, 'Scope')[0]
        link_title = LINK % ("uk-link-heading uk-text-meta", "%s.html" % valid_filename(doc).replace('.adoc', ''), "", title)
        link_category = LINK % ("uk-link-heading uk-text-meta", "Category_%s.html" % valid_filename(category), "", category)
        link_scope = LINK % ("uk-link-heading uk-text-meta", "Scope_%s.html" % valid_filename(scope), "", scope)

        timestamp = self.srvdtb.get_doc_timestamp(doc)
        fuzzy_date = fuzzy_date_from_timestamp(timestamp)

        tooltip ="%s" % (title)
        return DOC_CARD % (title, tooltip, link_title, timestamp, fuzzy_date, link_scope)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
# File: srv_utils.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: Generic functions module
"""

import os
import re
import glob
import json
import time
import math
import shutil
import random
import hashlib
import operator
import threading
import subprocess
import traceback as tb
from datetime import date, datetime
from kb4it.core.env import LPATH, GPATH, EOHMARK, FILE
from kb4it.core.log import get_logger

log = get_logger('KB4ITUtil')


def load_kbdict(source_path):
    """C0111: Missing function docstring (missing-docstring)."""
    source_path = valid_filename(source_path)
    KB4IT_DB_FILE = os.path.join(LPATH['DB'], 'kbdict-%s.json' % source_path)
    try:
        with open(KB4IT_DB_FILE, 'r') as fkb:
            kbdict = json.load(fkb)
    except FileNotFoundError:
        kbdict = {}
    log.debug("Current kbdict entries: %d", len(kbdict))
    return kbdict

def save_kbdict(kbdict, path, name=None):
    """C0111: Missing function docstring (missing-docstring)."""
    if name is None:
        target_path = valid_filename(path)
        KB4IT_DB_FILE = os.path.join(LPATH['DB'], 'kbdict-%s.json' % target_path)
    else:
        KB4IT_DB_FILE = os.path.join(path, '%s.json' % name)

    with open(KB4IT_DB_FILE, 'w') as fkb:
        json.dump(kbdict, fkb)
        log.debug("KBDICT %s saved", KB4IT_DB_FILE)


def copy_docs(docs, target):
    """C0111: Missing function docstring (missing-docstring)."""
    for doc in docs:
        try:
            shutil.copy('%s' % doc, target)
            log.debug("%s copied to %s", doc, target)
        except:
            log.warning("%s not found", doc)
    log.debug("%d documents copied to '%s'", len(docs), target)


def copydir(source, dest):
    """Copy a directory structure overwriting existing files.

    https://gist.github.com/dreikanter/5650973#gistcomment-835606
    """
    for root, dirs, files in os.walk(source):
        if not os.path.isdir(root):
            os.makedirs(root)

        for file in files:
            rel_path = root.replace(source, '').lstrip(os.sep)
            dest_path = os.path.join(dest, rel_path)

            if not os.path.isdir(dest_path):
                os.makedirs(dest_path)

            try:
                shutil.copyfile(os.path.join(root, file), os.path.join(dest_path, file))
            except PermissionError:
                log.warning("Check permissions for file: %s", file)


def get_source_docs(path):
    """C0111: Missing function docstring (missing-docstring)."""
    if path[:-1] != os.path.sep:
        path = path + os.path.sep

    pattern = os.path.join(path) + '*.adoc'
    docs = glob.glob(pattern)
    docs.sort(key=lambda y: y.lower())
    log.debug("\tFound %d asciidoctor documents", len(docs))

    return docs


def get_traceback():
    """Get traceback."""
    return tb.format_exc()


def exec_cmd(data):
    """Execute an operating system command.

    Return:
    - document
    - True if success, False if not
    - res is the output
    """
    doc, cmd, res = data
    process = subprocess.Popen([cmd], shell=True, stdout=subprocess.PIPE)
    outs, errs = process.communicate()
    if errs is None:
        return doc, True, res
    else:
        log.debug("Compiling %s: Error: %s", doc, errs)
        return doc, False, res


def set_max_frequency(dkeyurl):
    """Calculate and set max frequency."""
    max_frequency = 1
    for keyword in dkeyurl:
        cur_frequency = len(dkeyurl[keyword])
        if cur_frequency > max_frequency:
            max_frequency = cur_frequency

    return max_frequency


def get_font_size(frequency, max_frequency):
    """Get font size for a word based in its frequency."""
    proportion = int(math.log((frequency * 100) / max_frequency))

    if proportion < 1:
        size = 8
    elif proportion in range(1, 2):
        size = 10
    elif proportion in range(2, 3):
        size = 18
    elif proportion in range(3, 4):
        size = 36
    elif proportion in range(4, 5):
        size = 72
    else:
        size = 72

    return size


def extract_toc(source):
    """C0111: Missing function docstring (missing-docstring)."""
    toc = ''
    items = []
    lines = source.split('\n')
    s = e = n = 0

    for line in lines:
        if line.find("toctitle") > 0:
            s = n + 1
        if s > 0:
            if line.startswith('</div>') and n > s:
                e = n
                break
        n = n + 1

    if s > 0 and e > s:
        for line in lines[s:e]:
            if line.startswith('<li><a href='):
                modifier = """<li><a class="uk-link-heading" """
                line = line.replace("<li><a ", modifier)
            else:
                line = line.replace("sectlevel1", "uk-nav uk-nav-default")
                line = line.replace("sectlevel2", "uk-nav-sub")
                line = line.replace("sectlevel3", "uk-nav-sub")
                line = line.replace("sectlevel4", "uk-nav-sub")
            items.append(line)
        toc = '\n'.join(items)
    return toc

def delete_target_contents(target_path):
    """C0111: Missing function docstring (missing-docstring)."""
    if os.path.exists(target_path):
        if os.path.isdir(target_path):
            for file_object in os.listdir(target_path):
                file_object_path = os.path.join(target_path, file_object)
                if os.path.isfile(file_object_path):
                    os.unlink(file_object_path)
                else:
                    shutil.rmtree(file_object_path)
            log.debug("Contents of directory '%s' deleted successfully", target_path)
        elif os.path.isfile(target_path):
            os.unlink(target_path)
            log.debug("File '%s' deleted successfully", target_path)

def delete_files(files):
    for path in files:
        try:
            if os.path.exists(path):
                os.unlink(path)
        except FileNotFoundError as error:
            log.warning(error)
            log.warning(files)

def get_asciidoctor_attributes(docpath):
    """C0111: Missing function docstring (missing-docstring)."""
    props = {}
    try:
        # Get lines
        line = open(docpath, 'r').readlines()

        # Add document title (first line) to graph
        title = line[0][2:-1]
        props['Title'] = [title]

        # read the rest of properties until watermark
        for n in range(1, len(line)):
            if line[n].startswith(':'):
                key = line[n][1:line[n].find(':', 1)]
                values = line[n][len(key)+2:-1].split(',')
                props[key] = [value.strip() for value in values]
            elif line[n].startswith(EOHMARK):
                # Stop processing if EOHMARK is found
                break
    except Exception as error:
        log.error(error)
        log.error("Document %s could not be processed" % docpath)
    return props

def get_hash_from_dict(adict):
    """Get the SHA256 hash for a given dictionary."""
    alist = []
    for key in adict:
        alist.append(key.lower())
        elements = adict[key]
        for elem in elements:
            alist.append(elem.lower())
    alist.sort(key=lambda y: y.lower())
    string = ' '.join(alist)
    m = hashlib.sha256()
    m.update(string.encode())
    return m.hexdigest()

def valid_filename(s):
    """Return the given string converted to a string that can be used for a clean filename.

    Remove leading and trailing spaces; convert other spaces to
    underscores; and remove anything that is not an alphanumeric, dash,
    underscore, or dot.
    >>> valid_filename("john's portrait in 2004.jpg")
    'johns_portrait_in_2004.jpg'

    Borrowed from:
    https://github.com/django/django/blob/master/django/utils/text.py
    """
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)

def guess_datetime(sdate):
    """Guess datetime for a given string and return a normalized datetime"""

    found = False
    patterns = ["%d/%m/%Y", "%d/%m/%Y %H:%M", "%d/%m/%Y %H:%M:%S",
                "%d.%m.%Y", "%d.%m.%Y %H:%M", "%d.%m.%Y %H:%M:%S",
                "%d-%m-%Y", "%d-%m-%Y %H:%M", "%d-%m-%Y %H:%M:%S",
                "%Y/%m/%d", "%Y/%m/%d %H:%M", "%Y/%m/%d %H:%M:%S",
                "%Y.%m.%d", "%Y.%m.%d %H:%M", "%Y.%m.%d %H:%M:%S",
                "%Y-%m-%d", "%Y-%m-%d %H:%M", "%Y-%m-%d %H:%M:%S",
                "%Y/%m/%d", "%Y/%m/%d %H:%M", "%Y/%m/%d %H:%M:%S.%f",
                "%Y.%m.%d", "%Y.%m.%d %H:%M", "%Y.%m.%d %H:%M:%S.%f",
                "%Y-%m-%d", "%Y-%m-%d %H:%M", "%Y-%m-%d %H:%M:%S.%f",
               ]
    for pattern in patterns:
        if not found:
            try:
                td = datetime.strptime(sdate, pattern)
                ts = td.strftime("%Y-%m-%d %H:%M:%S")
                timestamp = datetime.strptime(ts, "%Y-%m-%d %H:%M:%S")
                found = True
            except ValueError:
                timestamp = None
    return timestamp

def file_timestamp(filename):
    """Return last modification datetime normalized of a file"""
    t = os.path.getmtime(filename)
    sdate = datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S")
    return sdate

def string_timestamp(string):
    dt = guess_datetime(string)
    sdate = datetime.fromtimestamp(dt).strftime("%Y-%m-%d %H:%M:%S")
    return sdate

def get_human_datetime(dt):
    """Return datetime for humans."""
    return "%s" % dt.strftime("%a, %b %d, %Y at %H:%M")

def fuzzy_date_from_timestamp(timestamp):
    # FIXME: Improve delta fuzzy dates
    """C0111: Missing function docstring (missing-docstring)."""
    if type(timestamp) == str:
        d1 = guess_datetime(timestamp)
    else:
        d1 = timestamp
    d2 = datetime.now()
    rdate = d2 - d1
    if rdate.days > 0:
        if rdate.days <= 31:
            return "%d days ago" % int(rdate.days)

        if rdate.days > 31 and rdate.days < 365:
            return "%d months ago" % int((rdate.days/31))

        if rdate.days >= 365:
            years = int(rdate.days/365)
            months = int( (rdate.days / years - 365) / 30 )
            if months > 0:
                return "%d years and %d months ago" % (years, months)
            else:
                if years > 1:
                    return "%d years ago" % years
                else:
                    return "1 year ago"

    hours = rdate.seconds / 3600
    if int(hours) > 0:
        return "%d hours ago" % int(hours)

    minutes = rdate.seconds / 60
    if int(minutes) > 0:
        return "%d minutes ago" % int(minutes)

    if int(rdate.seconds) > 0:
        return "%d seconds ago" % int(rdate.seconds)

    if int(rdate.seconds) == 0:
        return "Right now"

def sort_dictionary(adict, reverse=True):
    return sorted(adict.items(), key=operator.itemgetter(1), reverse=reverse)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Server module.

# File: theme.py
# Author: Tomás Vírseda
# License: GPL v3
# Description: snippet theme scripts
"""

import os

from kb4it.services.builder import KB4ITBuilder
from kb4it.core.util import valid_filename

class Theme(KB4ITBuilder):
    def hello(self):
        self.log.debug("This is the theme snippets")

    def generate_sources(self):
        pass

    def build(self):
        # Default pages
        self.create_page_properties()
        self.create_page_stats()
        self.create_page_index_all()
        self.create_page_index()
        self.create_page_about_app()
        self.create_page_about_theme()
        self.create_page_about_kb4it()
        self.create_page_help()

    def create_page_index(self):
        html = self.template('PAGE_INDEX')
        html_key = self.create_page_key_body('Module')
        source_dir = self.srvapp.get_source_path()
        lang = os.path.basename(source_dir)
        content = html % (lang, html_key)
        self.distribute('index', content)


    def create_metadata_section(self, doc):
        """Return a html block for displaying core and custom keys."""
        try:
            doc_path = os.path.join(self.srvapp.get_source_path(), doc)
            html = self.template('METADATA_SECTION_HEADER')
            ROW_CUSTOM_PROP = self.template('METADATA_ROW_CUSTOM_PROPERTY')
            ROW_CUSTOM_PROP_TIMESTAMP = self.template('METADATA_ROW_CUSTOM_PROPERTY_TIMESTAMP')
            custom_keys = self.srvdtb.get_custom_keys(doc)
            custom_props = ''
            for key in custom_keys:
                try:
                    values = self.get_html_values_from_key(doc, key)
                    labels = self.get_labels(values)
                    custom_props += ROW_CUSTOM_PROP % (valid_filename(key), key, labels)
                except Exception as error:
                    self.log.error("Key[%s]: %s", key, error)
            num_custom_props = len(custom_props)
            if  num_custom_props > 1:
                html += custom_props
            html += self.template('METADATA_SECTION_FOOTER')
        except Exception as error:
            msgerror = "%s -> %s" % (doc, error)
            self.log.error("\t\t%s", msgerror)
            html = ''
            raise

        return html

    def create_page_key_body(self, key):
        """Create key page."""
        source_dir = self.srvapp.get_source_path()
        values = self.srvdtb.get_all_values_for_key(key)
        num_values = len(values)
        html = self.template('BODY_KEY')

        # TAB Cloud
        cloud = self.create_tagcloud_from_key(key)

        # TAB Stats
        stats = ""
        leader_row = self.template('LEADER_ROW')
        for value in values:
            docs = self.srvdtb.get_docs_by_key_value(key, value)
            tpl_value_link = self.template('LEADER_ROW_VALUE_LINK')
            value_link = tpl_value_link % (valid_filename(key), valid_filename(value), value)
            stats += leader_row % (value_link, len(docs))

        return html % (cloud, stats)

    def get_doc_card(self, doc):
        source_dir = self.srvapp.get_source_path()
        DOC_CARD = self.template('CARD_DOC')
        DOC_CARD_FOOTER = self.template('CARD_DOC_FOOTER')
        LINK = self.template('LINK')
        title = self.srvdtb.get_values(doc, 'Title')[0]
        category = self.srvdtb.get_values(doc, 'Category')[0]
        scope = self.srvdtb.get_values(doc, 'Scope')[0]
        tags = self.srvdtb.get_values(doc, 'Tag')
        link_title = LINK % ("uk-link-heading uk-text-meta", "%s.html" % valid_filename(doc).replace('.adoc', ''), "", title)
        if len(category) > 0 and len(scope) > 0:
            link_category = LINK % ("uk-link-heading uk-text-meta", "Category_%s.html" % valid_filename(category), "", category)
            link_scope = LINK % ("uk-link-heading uk-text-meta", "Scope_%s.html" % valid_filename(scope), "", scope)
            footer = DOC_CARD_FOOTER % (link_category, link_scope)
        else:
            footer = ''

        tooltip ="%s" % (title)
        return DOC_CARD % (tooltip, link_title, '', footer)



import calendar
from calendar import HTMLCalendar
from datetime import datetime
from kb4it.core.service import Service
from kb4it.core.util import valid_filename, guess_datetime

class EventsCalendar(Service, HTMLCalendar):
    """Credit to: https://github.com/garthhumphreys/How-to-Use-Python-To-Create-A-Beautiful-Web-Calendar"""

    def initialize(self):
        super(HTMLCalendar, self).__init__(calendar.MONDAY)
        self.current_year = None
        self.ml = {}
        self.get_services()
        print

    def set_events_days(self, events_days):
        """
        Attach the list of event days as a property, so we can access it
        anywhere.
        """
        self.events_days = events_days #
        for i in range(1,13):
            self.ml[i] = False
        for month, day in events_days:
            self.ml[month] = True

    def set_events_docs(self, docs):
        self.events_docs = docs

    def set_events_html(self, html):
        self.events_html = html

    def get_services(self):
        self.srvbld = self.get_service('Builder')

    def formatday(self, day, weekday):
        """Return a day as a table cell."""
        eday = 0 # var for checking if it's a event day
        cal_date = (self.month, day) # create a tuple of the calendar month and day
        EVENTCAL_TD_NODAY = self.srvbld.template('EVENTCAL_TD_NODAY')
        EVENTCAL_TD_DAY_LINK = self.srvbld.template('EVENTCAL_TD_DAY_LINK')
        EVENTCAL_TD_DAY_NOLINK = self.srvbld.template('EVENTCAL_TD_DAY_NOLINK')
        if cal_date in self.events_days: # check if current calendar tuple date exist in our list of events days
            eday = day # if it does exist set the event day var with it
        if day == 0:
            return EVENTCAL_TD_NODAY # day outside month
        elif day == eday: # check if this is one of the events days, then change the
            EVENT_PAGE = "events_%4d%02d%02d" % (self.current_year, self.month, day)
            EVENT_PAGE_VALID_FNAME = valid_filename(EVENT_PAGE)
            return EVENTCAL_TD_DAY_LINK % (self.cssclasses[weekday], EVENT_PAGE_VALID_FNAME, day)
        else:
            return EVENTCAL_TD_DAY_NOLINK % (self.cssclasses[weekday], day)

    def formatweek(self, theweek):
        """Return a complete week as a table row."""
        EVENTCAL_TR_WEEK = self.srvbld.template('EVENTCAL_TR_WEEK')
        s = ''.join(self.formatday(d, wd) for (d, wd) in theweek)
        return EVENTCAL_TR_WEEK % s

    def formatweekheader(self):
        """Return a header for a week as a table row."""
        EVENTCAL_TR_WEEK_HEADER = self.srvbld.template('EVENTCAL_TR_WEEK_HEADER')
        s = ''.join(self.formatweekday(i) for i in self.iterweekdays())
        return EVENTCAL_TR_WEEK_HEADER % s

    def formatweekday(self, nd):
        day_format = {}
        day_format[0] = ("mon", "Monday", "M")
        day_format[1] = ("tue", "Tuesday", "T")
        day_format[2] = ("wed", "Wednesday", "W")
        day_format[3] = ("thu", "Thursday", "T")
        day_format[4] = ("fri", "Friday", "F")
        day_format[5] = ("sat", "Saturday", "S")
        day_format[6] = ("sun", "Sunday", "S")
        return """<th class="%s"><a class="uk-link-heading uk-text-bolder uk-link-muted" href="#" uk-tooltip="%s">%s</a></th>""" % day_format[nd]

    def formatmonthname(self, theyear, themonth, withyear=True):
        """Return a formatted month name."""
        LINK = self.srvbld.template('LINK')
        dt = guess_datetime("%4d-%02d-01" % (theyear, themonth))
        if withyear:
            month_name = datetime.strftime(dt, "%B %Y")
        else:
            month_name = datetime.strftime(dt, "%B")

        if self.ml[themonth]:
            link = LINK % ("uk-link-heading", "events_%4d%02d.html" % (theyear, themonth), "uk-text-uppercase uk-text-muted", month_name)
        else:
            # ~ link = LINK % ("uk-link-heading", "", "", month_name)
            link = """<span class="%s">%s</span>""" % ("uk-text-uppercase uk-text-muted", month_name)
        return link

    def formatmonth(self, theyear, themonth, withyear=False):
        """Override in order to add the month as a property."""
        self.month = themonth
        month = super(EventsCalendar, self).formatmonth(theyear, themonth, withyear=False)
        return month

    def formatyear(self, theyear, width=4):
        """Return a formatted year as a table of tables."""
        EVENTCAL_TABLE_YEAR = self.srvbld.template('EVENTCAL_TABLE_YEAR')
        EVENTCAL_TABLE_YEAR_TR_MONTHROW = self.srvbld.template('EVENTCAL_TABLE_YEAR_TR_MONTHROW')
        EVENTCAL_TABLE_YEAR_TD_MONTH = self.srvbld.template('EVENTCAL_TABLE_YEAR_TD_MONTH')
        v = []
        width = max(width, 1)
        MONTHS = ''
        January = 1
        for i in range(January, January+12, width):
            months = range(i, min(i+width, 13))
            TD_YEAR_MONTH = ''
            for m in months:
                TD_YEAR_MONTH += EVENTCAL_TABLE_YEAR_TD_MONTH % self.formatmonth(theyear, m, withyear=False)
            MONTHS += EVENTCAL_TABLE_YEAR_TR_MONTHROW % TD_YEAR_MONTH
        v.append(EVENTCAL_TABLE_YEAR % MONTHS)
        return ''.join(v)

    def formatyearpage(self, theyear, width=4, css='calendar.css', encoding=None):
        """Return a formatted year as a complete HTML page."""
        self.current_year = theyear
        return self.formatyear(theyear, width)

    def build_year_pagination(self, years):
        EVENTCAL_YEAR_PAGINATION = self.srvbld.template('EVENTCAL_YEAR_PAGINATION')
        EVENTCAL_YEAR_PAGINATION_ITEM = self.srvbld.template('EVENTCAL_YEAR_PAGINATION_ITEM')
        ITEMS = ''
        for yp in sorted(years):
            ITEMS += EVENTCAL_YEAR_PAGINATION_ITEM % (yp, yp)
        return EVENTCAL_YEAR_PAGINATION % ITEMS



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Setup KB4IT project.

# File: setup.py.
# Author: Tomás Vírseda
# License: GPL v3
# Description: setup.py tells you that the module/package you are about
# to install has been packaged and distributed with Distutils, which is
# the standard for distributing Python Modules.
"""

import os
import glob
from setuptools import setup

from kb4it.core.env import APP

with open('kb4it/resources/common/appdata/VERSION', 'r') as fver:
    VERSION = fver.read()

with open('kb4it/resources/common/appdata/pypi/README.rst', 'r') as f:
    LONG_DESCRIPTION = f.read()


def add_data(root_data):
    """Add data files from a given directory."""
    dir_files = []
    resdirs = set()
    for root, dirs, files in os.walk(root_data):
        resdirs.add(os.path.realpath(root))

    resdirs.remove(os.path.realpath(root_data))

    for directory in resdirs:
        files = glob.glob(directory+'/*')
        relfiles = []
        for thisfile in files:
            if not os.path.isdir(thisfile):
                relfiles.append(os.path.relpath(thisfile))

        num_files = len(files)
        if num_files > 0:
            dir_files.append((os.path.relpath(directory), relfiles))

    return dir_files


DATA_FILES = add_data('kb4it/resources')

setup(
    name=APP['shortname'],
    version=APP['version'],
    author=APP['author'],
    author_email=APP['author_email'],
    url=APP['website'],
    description='A static website generator based on Asciidoctor sources.',
    long_description=LONG_DESCRIPTION,
    download_url='https://github.com/t00m/KB4IT/archive/master.zip',
    license=APP['license'],
    packages=['kb4it', 'kb4it.core', 'kb4it.services'],
    # distutils does not support install_requires, but pip needs it to be
    # able to automatically install dependencies
    install_requires=[],
    include_package_data=True,
    data_files=DATA_FILES,
    zip_safe=False,
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Topic :: Documentation',
        'Topic :: Software Development :: Documentation',
        'Topic :: Utilities'
    ],
    entry_points={
        'console_scripts': [
            'kb4it = kb4it.kb4it:main',
            ],
        },
)
